# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from lmi.shell.compat import *


class LMIClassCacheEntry(object):
    """
    Class used for storing :py:class:`wbem.CIMClass` in
    :py:class:`.LMIShellCache`.

    :param CIMClass cim_class: :py:class:`wbem.CIMClass` to cache
    :param bool full_fetch: True, if class is cached with qualifiers
    """
    def __init__(self, cim_class, full_fetch):
        self.cim_class = cim_class
        self.full_fetch = full_fetch


class LMIShellCache(object):
    """
    Class representing a LMIShell cache.

    :param bool active: specifies, if the cache is active
    :param list classname_list: list of strings of cached class names
    :param dictionary class_dict: cached :py:class:`wbem.CIMClass` objects,
        where the key is the class name and value is :class:`CIMClass` object
    :param dictionary class_superclass_dict: dictionary, where the key is
        namespace and value is dictionary of classname:superclass
    """
    def __init__(self, active=True, classname_dict=None, class_dict=None,
                 class_superclass_dict=None):
        self._classname_dict = classname_dict \
            if classname_dict is not None else {}
        self._class_dict = class_dict if class_dict is not None else {}
        self._class_superclass_dict = class_superclass_dict \
            if class_superclass_dict is not None else {}
        self._active = active

    def clear(self):
        """
        Clears the cache.
        """
        self._classname_dict = {}
        self._class_dict = {}
        self._class_superclass_dict = {}

    def get_classes(self, namespace=wbem.DEFAULT_NAMESPACE):
        """
        :param string namespace: namespace storing cached classes
        :returns: list of cached class names or None, if no cached
            classes is stored
        """
        return self._classname_dict.get(namespace, None)

    def set_classes(self, classname_list, namespace=wbem.DEFAULT_NAMESPACE):
        """
        Stores a new class names' list.

        :param string namespace: namespace storing cached classes
        """
        self._classname_dict[namespace] = classname_list

    def get_class(self, classname, namespace=wbem.DEFAULT_NAMESPACE):
        """
        :param string classname: cached class name
        :param string namespace: namespace storing cached classes
        :returns: cache object, if proper class name provided, None otherwise
        :rtype: :py:class:`.LMIClassCacheEntry`
        """
        if namespace not in self._class_dict or \
                classname not in self._class_dict[namespace]:
            return None
        return self._class_dict[namespace][classname]

    def add_class(self, cim_class, namespace=wbem.DEFAULT_NAMESPACE,
                  full_fetch=False):
        """
        Stores a new :py:class:`wbem.CIMClass` object into the cache.

        :param CIMClass cim_class: :py:class:`wbem.CIMClass` object
        :param string namespace: namespace storing cached classes
        """
        if namespace not in self._class_dict:
            self._class_dict[namespace] = {}
        self._class_dict[namespace][cim_class.classname] = LMIClassCacheEntry(
            cim_class, full_fetch)

    def has_superclass(self, classname, namespace):
        """
        :param string classname: cached class name
        :param string namespace: namespace name
        :returns: True, if the cache contains superclass to the given class
            name; False otherwise
        """
        if namespace not in self._class_superclass_dict:
            return False
        if classname not in self._class_superclass_dict[namespace]:
            return False
        return True

    def get_superclass(self, classname, namespace):
        """
        :param string classname: cached class name
        :param string namespace: namespace name
        :returns: cached superclass to the given class name
        :rtype: string
        """
        if not self.has_superclass(classname, namespace):
            return None
        return self._class_superclass_dict[namespace][classname]

    def add_superclass(self, classname, superclass, namespace):
        """
        Stores a new pair classname : superclassname into the cache.

        :param string classname: class name to be stored
        :param string superclass: super class name to be stored
        :param string namespace: namespace name of the classname
        """
        if namespace not in self._class_superclass_dict:
            self._class_superclass_dict[namespace] = {}
        self._class_superclass_dict[namespace][classname] = superclass

    @property
    def active(self):
        """
        :returns: True, if the cache is active; False otherwise
        """
        return self._active

    @active.setter
    def active(self, val):
        """
        Property setter for the property active.
        """
        self._active = val



# Copyright (C) 2013-2014 Red Hat, Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of the FreeBSD Project.
#
# Authors: Michal Minar <miminar@redhat.com>
#
"""
Parser for mini-language specifying profile and class requirements. We call
the language LMIReSpL (openLMI Requirement Specification Language).

The only thing designed for use outside this module is :py:func:`bnf_parser`.

Language is generated by BNF grammer which served as a model for parser.

Formal representation of BNF grammer is following: ::

    expr          ::= term [ op expr ]*
    term          ::= '!'? req
    req           ::= profile_cond | clsreq_cond | '(' expr ')'
    profile_cond  ::= 'profile'? [ profile | profile_quot ] cond?
    clsreq_cond   ::= 'class' [ clsname | clsname_quot] cond?
    profile_quot  ::= '"' /\w+[ +.a-zA-Z0-9_-]*/ '"'
    profile       ::= /\w+[+.a-zA-Z_-]*/
    clsname_quot  ::= '"' clsname '"'
    clsname       ::= /[a-zA-Z]+_[a-zA-Z][a-zA-Z0-9_]*/
    cond          ::= cmpop version
    cmpop         ::= /(<|=|>|!)=|<|>/
    version       ::= /[0-9]+(\.[0-9]+)*/
    op            ::= '&' | '|'

String surrounded by quotes is a literal. String enclosed with slashes is a
regular expression. Square brackets encloses a group of words and limit
the scope of some operation (like iteration).
"""
import abc
import operator
from pyparsing import Literal, Combine, Optional, ZeroOrMore, \
        Forward, Regex, Keyword, FollowedBy, LineEnd, ParseException

#: Dictionary mapping supported comparison operators to a pair. First item is a
#: function making the comparison and the second can be of two values (``all``
#: or ``any``). Former sayes that each part of first version string must be in
#: relation to corresponding part of second version string in order to satisfy
#: the condition. The latter causes the comparison to end on first satisfied
#: part.
OP_MAP = {
        '==' : (operator.eq, all),
        '<=' : (operator.le, all),
        '>=' : (operator.ge, all),
        '!=' : (operator.ne, any),
        '>'  : (operator.gt, any),
        '<'  : (operator.lt, any)
}

def cmp_version(fst, snd, opsign='<'):
    """
    Compare two version specifications. Each version string shall contain
    digits delimited with dots. Empty string is also valid version. It will be
    replaced with -1.

    :param str fst: First version string.
    :param str snd: Second version string.
    :param str opsign: Sign denoting operation to be used. Supported signs
        are present in :py:attr:`OP_MAP`.
    :returns: ``True`` if the relation denoted by particular operation exists
        between two operands.
    :rtype: boolean
    """
    def splitver(ver):
        """ Converts version string to a tuple of integers. """
        return tuple(int(p) if p else -1 for p in ver.split('.'))
    aparts = splitver(fst)
    bparts = splitver(snd)
    op, which = OP_MAP[opsign]
    if which is all:
        for ap, bp in zip(aparts, bparts):
            if not op(ap, bp):
                return False
    else:
        for ap, bp in zip(aparts, bparts):
            if op(ap, bp):
                return True
            if ap != bp:
                return False
    return op(len(aparts), len(bparts))

class SemanticGroup(object):
    """
    Base class for non-terminals. Just a minimal set of non-terminals is
    represented by objects the rest is represented by strings.

    All subclasses need to define their own :py:meth:`evaluate` method. The
    parser builds a tree of these non-terminals with single non-terminal being
    a root node. This node's *evaluate* method returns a boolean saying whether
    the condition is satisfied. Root node is always an object of
    :py:class:`Expr`.
    """

    __metaclass__ = abc.ABCMeta

    def __call__(self):
        return self.evaluate()

    @abc.abstractmethod
    def evaluate(self):
        """
        :returns: ``True`` if the sub-condition represented by this non-terminal
            is satisfied.
        :rtype: boolean
        """
        pass

class Expr(SemanticGroup):
    """
    Initial non-terminal. Object of this class (or one of its subclasses) is a
    result of parsing.

    :param term: An object of :py:class:`Term` non-terminal.
    """

    def __init__(self, term):
        assert isinstance(term, Term)
        self.fst = term

    def evaluate(self):
        return self.fst()

    def __str__(self):
        return str(self.fst)

class And(Expr):
    """
    Represents logical *AND* of two expressions. Short-circuit evaluation is
    being exploited here.

    :param fst: An object of :py:class:`Term` non-terminal.
    :param snd: An object of :py:class:`Term` non-terminal.
    """

    def __init__(self, fst, snd):
        assert isinstance(snd, (Term, Expr))
        Expr.__init__(self, fst)
        self.snd = snd

    def evaluate(self):
        if self.fst():
            return self.snd()
        return False

    def __str__(self):
        return "%s & %s" % (self.fst, self.snd)

class Or(Expr):
    """
    Represents logical *OR* of two expressions. Short-circuit evaluation is being
    exploited here.

    :param fst: An object of :py:class:`Term` non-terminal.
    :param snd: An object of :py:class:`Term` non-terminal.
    """

    def __init__(self, fst, snd):
        assert isinstance(snd, (Term, Expr))
        Expr.__init__(self, fst)
        self.snd = snd

    def evaluate(self):
        if self.fst():
            return True
        return self.snd()

    def __str__(self):
        return "%s | %s" % (self.fst, self.snd)

class Term(SemanticGroup):
    """
    Represents possible negation of expression.

    :param req: An object of :py:class:`Req`.
    :param boolean negate: Whether the result of children shall be negated.
    """

    def __init__(self, req, negate):
        assert isinstance(req, Req)
        self.req = req
        self.negate = negate

    def evaluate(self):
        res = self.req()
        return not res if self.negate else res

    def __str__(self):
        if self.negate:
            return '!' + str(self.req)
        return str(self.req)

class Req(SemanticGroup):
    """
    Represents one of following subexpressions:

        * single requirement on particular profile
        * single requirement on particular class
        * a subexpression
    """
    pass

class ReqCond(Req):
    """
    Represents single requirement on particular class or profile.

    :param str kind: Name identifying kind of thing this belongs. For example
        ``'class'`` or ``'profile'``.
    :param callable version_getter: Is a function called to get version of
        either profile or CIM class. It must return corresponding version string
        if the profile or class is registered and ``None`` otherwise.
        Version string is read from ``RegisteredVersion`` property of
        ``CIM_RegisteredProfile``. If a class is being queried, version
        shall be taken from ``Version`` qualifier of given class.
    :param str name: Name of profile or CIM class to check for. In case
        of a profile, it is compared to ``RegisteredName`` property of
        ``CIM_RegisteredProfile``. If any instance of this class has matching
        name, it's version will be checked. If no matching instance is found,
        instances of ``CIM_RegisteredSubProfile`` are queried the same way.
        Failing to find it results in ``False``.
    :param str cond: Is a version requirement. Check the grammer above for
        ``cond`` non-terminal.
    """

    def __init__(self, kind, version_getter, name, cond=None):
        assert isinstance(kind, basestring)
        assert callable(version_getter)
        assert isinstance(name, basestring)
        assert cond is None or (isinstance(cond, tuple) and len(cond) == 2)
        self.kind = kind
        self.version_getter = version_getter
        self.name = name
        self.cond = cond

    def evaluate(self):
        version = self.version_getter(self.name)
        return version and (not self.cond or self._check_version(version))

    def _check_version(self, version):
        """
        Checks whether the version of profile or class satisfies the
        requirement. Version strings are first split into a list of integers
        (that were delimited with a dot) and then they are compared in
        descending order from the most signigicant down.

        :param str version: Version of profile or class to check.
        """
        opsign, cmpver = self.cond
        return cmp_version(version, cmpver, opsign)

    def __str__(self):
        return '{%s "%s"%s}' % (
                self.kind, self.name, ' %s %s' % self.cond if self.cond else '')

class Subexpr(Req):
    """
    Represents a subexpression originally enclosed in brackets.
    """

    def __init__(self, expr):
        assert isinstance(expr, Expr)
        self.expr = expr

    def evaluate(self):
        return self.expr()

    def __str__(self):
        return "(%s)" % self.expr

class TreeBuilder(object):
    """
    A stack interface for parser. It defines methods modifying the stack with
    additional checks.
    """

    def __init__(self, stack, profile_version_getter, class_version_getter):
        if not isinstance(stack, list):
            raise TypeError("stack needs to be empty!")
        if stack:
            stack[:] = []
        self.stack = stack
        self.profile_version_getter = profile_version_getter
        self.class_version_getter = class_version_getter

    def expr(self, strg, loc, toks):
        """
        Operates upon a stack. It takes either one or two *terms* there
        and makes an expression object out of them. Terms need to be delimited
        with logical operator.
        """
        assert len(self.stack) > 0
        if not isinstance(self.stack[-1], (Term, Expr)):
            raise ParseException("Invalid expression (stopped at char %d)."
                    % loc)
        if len(self.stack) >= 3 and self.stack[-2] in ('&', '|'):
            assert isinstance(self.stack[-3], Term)
            if self.stack[-2] == '&':
                expr = And(self.stack[-3], self.stack[-1])
            else:
                expr = Or(self.stack[-3], self.stack[-1])
            self.stack.pop()
            self.stack.pop()
        elif not isinstance(self.stack[-1], Expr):
            expr = Expr(self.stack[-1])
        else:
            expr = self.stack[-1]
        self.stack[-1] = expr

    def term(self, strg, loc, toks):
        """
        Creates a ``term`` out of requirement (``req`` non-terminal).
        """
        assert len(self.stack) > 0
        assert isinstance(self.stack[-1], Req)
        self.stack[-1] = Term(self.stack[-1], toks[0] == '!')

    def subexpr(self, strg, loc, toks):
        """
        Operates upon a stack. It creates an instance of :py:class:`Subexpr`
        out of :py:class:`Expr` which is enclosed in brackets.
        """
        assert len(self.stack) > 1
        assert self.stack[-2] == '('
        assert isinstance(self.stack[-1], Expr)
        assert len(toks) > 0 and toks[-1] == ')'
        self.stack[-2] = Subexpr(self.stack[-1])
        self.stack.pop()

    def push_class(self, strg, loc, toks):
        """
        Handles ``clsreq_cond`` non-terminal in one go. It extracts
        corresponding tokens and pushes an object of :py:class:`ReqCond` to a
        stack.
        """
        assert toks[0] == 'class'
        assert len(toks) >= 2
        name = toks[1]
        condition = None
        if len(toks) > 2 and toks[2] in OP_MAP:
            assert len(toks) >= 4
            condition = toks[2], toks[3]
        self.stack.append(ReqCond('class', self.class_version_getter,
            name, condition))

    def push_profile(self, strg, loc, toks):
        """
        Handles ``profile_cond`` non-terminal in one go. It behaves in the same
        way as :py:meth:`push_profile`.
        """
        index = 0
        if toks[0] == 'profile':
            index = 1
        assert len(toks) > index
        name = toks[index]
        index += 1
        condition = None
        if len(toks) > index and toks[index] in OP_MAP:
            assert len(toks) >= index + 2
            condition = toks[index], toks[index + 1]
        self.stack.append(ReqCond('profile', self.profile_version_getter,
            name, condition))

    def push_literal(self, strg, loc, toks):
        """
        Pushes operators to a stack.
        """
        assert toks[0] in ('&', '|', '(')
        if toks[0] == '(':
            assert not self.stack or self.stack[-1] in ('&', '|')
        else:
            assert len(self.stack) > 0
            assert isinstance(self.stack[-1], Term)
        self.stack.append(toks[0])

def bnf_parser(stack, profile_version_getter, class_version_getter):
    """
    Builds a parser operating on provided stack.

    :param list stack: Stack to operate on. It will contain the resulting
        :py:class:`Expr` object when the parsing is successfully over -
        it will be the only item in the list. It needs to be initially empty.
    :param callable profile_version_getter: Function returning version
        of registered profile or ``None`` if not present.
    :param callable class_version_getter: Fucntion returning version
        of registered class or ``None`` if not present.
    :returns: Parser object.
    :rtype: :py:class:`pyparsing,ParserElement`
    """
    if not isinstance(stack, list):
        raise TypeError("stack must be a list!")
    builder = TreeBuilder(stack, profile_version_getter, class_version_getter)

    ntop = ((Literal('&') | Literal('|')) + FollowedBy(Regex('["a-zA-Z\(!]'))) \
            .setName('op').setParseAction(builder.push_literal)
    ntversion = Regex(r'[0-9]+(\.[0-9]+)*').setName('version')
    ntcmpop = Regex(r'(<|=|>|!)=|<|>(?=\s*\d)').setName('cmpop')
    ntcond = (ntcmpop + ntversion).setName('cond')
    ntclsname = Regex(r'[a-zA-Z]+_[a-zA-Z][a-zA-Z0-9_]*').setName('clsname')
    ntclsname_quot = Combine(
              Literal('"').suppress()
            + ntclsname
            + Literal('"').suppress()).setName('clsname_quot')
    ntprofile_quot = Combine(
              Literal('"').suppress()
            + Regex(r'\w+[ +.a-zA-Z0-9_-]*')
            + Literal('"').suppress()).setName('profile_quot')
    ntprofile = Regex(r'\w+[+.a-zA-Z0-9_-]*').setName('profile')
    ntclsreq_cond = (
              Keyword('class')
            + (ntclsname_quot | ntclsname)
            + Optional(ntcond)).setName('clsreq_cond').setParseAction(
                    builder.push_class)
    ntprofile_cond = (
              Optional(Keyword('profile'))
            + (ntprofile_quot | ntprofile)
            + Optional(ntcond)).setName('profile_cond').setParseAction(
                    builder.push_profile)
    ntexpr = Forward().setName('expr')
    bracedexpr = (
              Literal('(').setParseAction(builder.push_literal)
            + ntexpr
            + Literal(')')).setParseAction(builder.subexpr)
    ntreq = (bracedexpr | ntclsreq_cond | ntprofile_cond).setName('req')
    ntterm = (Optional(Literal("!")) + ntreq + FollowedBy(Regex('[\)&\|]') | LineEnd()))\
            .setParseAction(builder.term)
    ntexpr << ntterm + ZeroOrMore(ntop + ntexpr).setParseAction(builder.expr)

    return ntexpr

if __name__ == '__main__':
    def get_class_version(class_name):
        try:
            version = { 'lmi_logicalfile'      : '0.1.2'
                      , 'lmi_softwareidentity' : '3.2.1'
                      , 'pg_computersystem'    : '1.1.1'
                      }[class_name.lower()]
        except KeyError:
            version = None
        return version

    def get_profile_version(profile_name):
        try:
            version = { 'openlmi software' : '0.1.2'
                      , 'openlmi-software' : '1.3.4'
                      , 'openlmi hardware' : '1.1.1'
                      , 'openlmi-hardware' : '0.2.3'
                      }[profile_name.lower()]
        except KeyError:
            version = None
        return version

    def test(s, expected):
        stack = []
        parser = bnf_parser(stack, get_profile_version, get_class_version)
        results = parser.parseString(s, parseAll=True)
        if len(stack) == 1:
            evalresult = stack[0]()
            if expected == evalresult:
                print "%s\t=>\tOK" % s
            else:
                print "%s\t=>\tFAILED" % s
        else:
            print "%s\t=>\tFAILED" % s
        print "  stack: [%s]" % ', '.join(str(i) for i in stack)

    test( 'class LMI_SoftwareIdentity == 0.2.0', False)
    test( '"OpenLMI-Software" == 0.1.2 & "OpenLMI-Hardware" < 0.1.3', False)
    test( 'OpenLMI-Software<1|OpenLMI-Hardware!=1.2.4', True)
    test( '"OpenLMI Software" & profile "OpenLMI Hardware"'
          ' | ! class LMI_LogicalFile', True)
    test( 'profile OpenLMI-Software > 0.1.2 & !(class "PG_ComputerSystem"'
          ' == 2.3.4 | "OpenLMI Hardware")', False)
    test( 'OpenLMI-Software > 1.3 & OpenLMI-Software >= 1.3.4'
          ' & OpenLMI-Software < 1.3.4.1 & OpenLMI-Software <= 1.3.4'
          ' & OpenLMI-Software == 1.3.4', True)
    test( 'OpenLMI-Software < 1.3.4 | OpenLMI-Software > 1.3.4'
          ' | OpenLMI-Software != 1.3.4', False)
    test( '(! OpenLMI-Software == 1.3.4 | OpenLMI-Software <= 1.3.4.1)'
          ' & !(openlmi-software > 1.3.4 | Openlmi-software != 1.3.4)', True)
    for badexpr in (
            'OpenLMI-Software > & OpenLMI-Hardware',
            'classs LMI_SoftwareIdentity',
            'OpenLMI-Software > 1.2.3 | == 5.4.3',
            '',
            '"OpenLMI-Software',
            'OpenLMI-Software < > OpenLMI-Hardware',
            'OpenlmiSoftware & (openLMI-Hardware ',
            'OpenLMISoftware & ) OpenLMI-Hardare (',
            'OpenLMISoftware | OpenlmiSoftware > "1.2.3"'
            ):
        try:
            test(badexpr, None)
        except ParseException:
            print "%s\t=>\tOK" % badexpr



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from lmi.shell.compat import *

from lmi.shell.LMIObjectFactory import LMIObjectFactory

# NOTE: the enumeration comes from DMTF standard.
JOB_STATE_COMPLETED, \
    JOB_STATE_TERMINATED, \
    JOB_STATE_KILLED, \
    JOB_STATE_EXCEPTION = range(7, 11)

# Job finish states are used in
# LMIMethod.__handle_synchro_method_call_indication()
#    JOB_FINISH_DELAYED -- a job has completed its action later,
#       an indication or polling method was used
#    JOB_FINISHED_EARLY -- no waiting method was used
JOB_NOT_FINISHED, \
    JOB_FINISH_DELAYED, \
    JOB_FINISH_EARLY = range(3)


def _lmi_get_job_state(job):
    """
    Helper function, which returns a numeric Job state value.

    :param job: :py:class:`.LMIInstance` or :py:class:`wbem.CIMInstance`
        representing a job
    """
    if isinstance(job, LMIObjectFactory().LMIInstance):
        return job.JobState
    elif isinstance(job, wbem.CIMInstance):
        return job["JobState"]
    raise TypeError("'job' object needs to be instance of either LMIInstance "
                    "or CIMInstance")


def lmi_is_job_finished(job):
    """
    Helper function, which returns bool flag, if the job is in finished state.

    :param job: :py:class:`.LMIInstance` or :py:class:`wbem.CIMInstance`
        representing a job
    """
    return _lmi_get_job_state(job) in (
        JOB_STATE_COMPLETED,
        JOB_STATE_TERMINATED,
        JOB_STATE_KILLED,
        JOB_STATE_EXCEPTION)


def lmi_is_job_completed(job):
    """
    Helper function, which returns bool flag, if the job is in completed state.

    :param job: :py:class:`.LMIInstance` or :py:class:`wbem.CIMInstance`
        representing a job
    """
    return _lmi_get_job_state(job) == JOB_STATE_COMPLETED


def lmi_is_job_terminated(job):
    """
    Helper function, which returns bool flag, if the job is in terminated
    state.

    :param job: :py:class:`.LMIInstance` or :py:class:`wbem.CIMInstance`
        representing a job
    """
    return _lmi_get_job_state(job) == JOB_STATE_TERMINATED


def lmi_is_job_killed(job):
    """
    Helper function, which returns bool flag, if the job is killed.

    :param job: :py:class:`.LMIInstance` or :py:class:`wbem.CIMInstance`
        representing a job
    """
    return _lmi_get_job_state(job) == JOB_STATE_KILLED


def lmi_is_job_exception(job):
    """
    Helper function, which returns bool flag, if the job is in the exception
    state.

    :param job: :py:class:`.LMIInstance` or :py:class:`wbem.CIMInstance`
        representing a job
    """
    return _lmi_get_job_state(job) == JOB_STATE_EXCEPTION



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import keyword
import os.path
import rlcompleter

from lmi.shell.LMIReturnValue import LMIReturnValue
from lmi.shell.LMIConstantValues import LMIConstantValues
from lmi.shell.LMINamespace import LMINamespace
from lmi.shell.LMINamespace import LMINamespaceRoot
from lmi.shell.LMIClass import LMIClass
from lmi.shell.LMIInstance import LMIInstance
from lmi.shell.LMIInstanceName import LMIInstanceName
from lmi.shell.LMIMethod import LMIMethod
from lmi.shell.LMIConnection import LMIConnection


class LMICompleter(rlcompleter.Completer):
    """
    This LMIShell completer, which is used in the interactive mode, provides
    tab-completion for user friendliness.

    :param dictionary namespace: dictionary, where to perform a completion. If
        unspecified, the default namespace where completions are performed is
        :samp:`__main__` (technically, :samp:`__main__.__dict__`).
    """
    def __init__(self, namespace=None):
        rlcompleter.Completer.__init__(self, namespace)

    def _callable_postfix(self, val, word):
        """
        :param val: object, which is checked, if it is callable
        :param string word: input string
        :returns: string with opening parentheses, if the value is callable
        """
        if hasattr(val, word) and callable(getattr(val, word)):
            word = word + "("
        return word

    def complete(self, text, state):
        """
        :param string text: string to be completed.
        :param state: order number of the completion, see rlcompleter
        :returns: completed string
        """
        if not text:
            return ("\t", None)[state]
        return rlcompleter.Completer.complete(self, text, state)

    def global_matches(self, text):
        """
        :param string text: expression to complete
        :returns: list of all keywords, built-in functions and names
        :rtype: list of strings
        """
        names = rlcompleter.Completer.global_matches(self, text)
        prefix = os.path.commonprefix(names)
        if prefix and prefix != text:
            return [prefix]
        return names

    def attr_matches(self, text):
        """
        :param string text: expression to complete
        :returns: list of attributes of a given expression; if the expression
            is instance of LMI wrapper class, its important
            properties/attributes/ methods/parameters will be added too
        :rtype: list of strings
        """
        expr, attr = text.rsplit(".", 1)
        if "(" in expr or ")" in expr:
            # Do not evaluate a callable
            return

        obj = eval(expr, self.namespace)
        words = []
        if isinstance(obj, LMINamespace):
            words.extend([cls for cls in obj.classes() if cls[0] != "_"])
            if isinstance(obj, LMINamespaceRoot):
                words.extend(obj.namespaces)
        elif isinstance(obj, LMIClass):
            words.extend([p + "Values" for p in obj.valuemap_properties()])
        elif isinstance(obj, LMIInstance):
            words.extend([m + "(" for m in obj.methods()])
            words.extend(obj.properties())
        elif isinstance(obj, LMIInstanceName):
            words.extend([m + "(" for m in obj.methods()])
            words.extend(obj.key_properties())
        elif isinstance(obj, LMIMethod):
            words.extend([p + "Values" for p in obj.valuemap_parameters()])
        elif isinstance(obj, LMIConstantValues):
            words.extend(obj.values())
        elif isinstance(obj, LMIReturnValue):
            words.extend(obj.properties())
        words.extend([self._callable_postfix(obj, member)
                      for member in dir(obj) if member[0] != "_"])

        # Search for words, which can complete the current expression.
        names = filter(lambda word: word.startswith(attr), words)

        # Search for common prefix of all available words we currently got. If
        # there is any, return such completion.
        prefix = os.path.commonprefix(names)
        if prefix and prefix != attr:
            return ["%s.%s" % (expr, prefix)]

        # Return all the possible completions or an empty list, if there is no
        # completion available (to prevent rlcompleter to erase the whole
        # text).
        return names + [" "] if names else []



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import sys
import subprocess
from site import _Helper


class LMIHelper(_Helper):
    """
    LMI Helper class, which overrides python help.
    """
    def __repr__(self):
        """
        :returns: pretty string for the object.
        """
        return "Type help() to see man page for lmishell, " \
            "or help(object) for help about object."

    def __call__(self, request=None):
        """
        Executes either man page for LMIShell or prints pydoc help for an
        object.
        """
        if request is None:
            try:
                # Python 3+
                from subprocess import DEVNULL as devnull
                devnull_close = lambda: None
            except ImportError:
                devnull = open(os.devnull, "wb")
                devnull_close = lambda: devnull.close()
            rcode = subprocess.call(["man", "lmishell"], stderr=devnull)
            devnull_close()
            if rcode > 0:
                sys.stderr.write("Man page for lmishell can not be found, ")
                sys.stderr.write("refer to official documentation.\n\n")
                sys.stderr.write("Available at: http://www.openlmi.org\n")
        else:
            import pydoc
            return pydoc.help(request)



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Defines command classes producing tablelike output.
"""

import abc
from itertools import chain

from lmi.scripts.common import errors
from lmi.scripts.common import formatter
from lmi.scripts.common import get_logger
from lmi.scripts.common.command import meta
from lmi.scripts.common.command.session import LmiSessionCommand
from lmi.scripts.common.configuration import Configuration
from lmi.scripts.common.formatter import command as fcmd

LOG = get_logger(__name__)

class LmiBaseListerCommand(LmiSessionCommand):
    """
    Base class for all lister commands.
    """

    @classmethod
    def get_columns(cls):
        """
        :returns: Column names for resulting table. ``COLUMNS`` property
            will be converted to this class method. If ``None``, the associated
            function shall return column names as the first tuple of returned
            list. If empty tuple or list, no header shall be printed and associated
            function returns just data rows.
        :rtype: list or tuple or None
        """
        return None

    def formatter_factory(self):
        if self.app.config.lister_format == Configuration.LISTER_FORMAT_CSV:
            return formatter.CsvFormatter
        else:
            return formatter.TableFormatter

class LmiLister(LmiBaseListerCommand):
    """
    End point command outputting a table for each host. Associated function
    shall return a list of rows. Each row is represented as a tuple holding
    column values.

    List of additional recognized properties:

        ``COLUMNS`` : ``tuple``
            Column names. It's a tuple with name for each column. Each row
            shall then contain the same number of items as this tuple. If
            omitted, associated function is expected to provide them in the
            first row of returned list. It's translated to ``get_columns()``
            class method.

    Using metaclass: :py:class:`~.meta.ListerMetaClass`.
    """
    __metaclass__ = meta.ListerMetaClass

    def take_action(self, connection, args, kwargs):
        """
        Collects results of single host.

        :param connection:  Connection to a single host.
        :type connection: :py:class:`lmi.shell.LMIConnection`
        :param list args: Positional arguments for associated function.
        :param dictionary kwargs: Keyword arguments for associated function.
        :returns: Column names and item list as a pair.
        :rtype: tuple
        """
        res = self.execute_on_connection(connection, *args, **kwargs)
        columns = self.get_columns()
        if isinstance(columns, (tuple, list)) and columns:
            command = fcmd.NewTableHeaderCommand(columns)
            res = chain((command, ), res)
        elif columns is None:
            resi = iter(res)
            command = fcmd.NewTableHeaderCommand(resi.next())
            res = chain((command, ), resi)
        return res

class LmiInstanceLister(LmiBaseListerCommand):
    """
    End point command outputting a table of instances for each host.
    Associated function shall return a list of instances. They may be
    prepended with column names depending on value of ``DYNAMIC_PROPERTIES``.
    Each instance will occupy single row of table with property values being a
    content of cells.

    List of additional recognized properties is the same as for
    :py:class:`~.show.LmiShowInstance`. There is just one difference. Either
    ``DYNAMIC_PROPERTIES`` must be ``True`` or ``PROPERTIES`` must be filled.

    Using metaclass: :py:class:`~.meta.InstanceListerMetaClass`.
    """
    __metaclass__ = meta.InstanceListerMetaClass

    @abc.abstractmethod
    def render(self, result):
        """
        This method can either be overriden in a subclass or left alone. In the
        latter case it will be generated by
        :py:class:`~.meta.InstanceListerMetaClass` metaclass with regard to
        ``PROPERTIES`` and ``DYNAMIC_PROPERTIES``.

        :param result: Either an instance to render or pair of properties and
            instance.
        :type result: :py:class:`lmi.shell.LMIInstance` or tuple
        :returns: List of pairs, where the first item is a label and second a
            value to render.
        :rtype: list
        """
        raise NotImplementedError(
                "render method must be overriden in subclass")

    def take_action(self, connection, args, kwargs):
        """
        Collects results of single host.

        :param connection: Connection to a single host.
        :type connection: :py:class:`lmi.shell.LMIConnection`
        :param list args: Positional arguments for associated function.
        :param dictionary kwargs: Keyword arguments for associated function.
        :returns: Column names and item list as a pair.
        :rtype: tuple
        """
        cols = self.get_columns()
        if cols is None:
            result = self.execute_on_connection(
                    connection, *args, **kwargs)
            if not isinstance(result, tuple) or len(result) != 2:
                raise errors.LmiUnexpectedResult(
                        self.__class__, "(properties, instances)", result)
            cols, data = result
            if not isinstance(cols, (tuple, list)):
                raise errors.LmiUnexpectedResult(
                        self.__class__, "(tuple, ...)", (cols, '...'))
            header = [c if isinstance(c, basestring) else c[0] for c in cols]
            cmd = fcmd.NewTableHeaderCommand(columns=header)
            return chain((cmd, ), (self.render((cols, inst)) for inst in data))
        else:
            data = self.execute_on_connection(connection, *args, **kwargs)
            if not hasattr(data, '__iter__'):
                raise errors.LmiUnexpectedResult(
                        self.__class__, 'list or generator', data)
            cmd = fcmd.NewTableHeaderCommand(columns=cols)
            return chain((cmd, ), (self.render(inst) for inst in data))



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
LMIShell
"""
import sys

# Setup logging format
from lmi.shell.LMIShellLogger import lmi_init_logger
lmi_init_logger()

from lmi.shell.LMIShellVersion import __version__

from lmi.shell.compat import *
from lmi.shell.LMIExceptions import *
from lmi.shell.LMIUtil import LMIPassByRef
from lmi.shell.LMIIndicationListener import LMIIndicationListener
from lmi.shell.LMIReturnValue import LMIReturnValue
from lmi.shell.LMICIMXMLClient import LMICIMXMLClient
from lmi.shell.LMIWSMANClient import LMIWSMANClient
from lmi.shell.LMIShellClient import LMIShellClient
from lmi.shell.LMIShellConfig import LMIShellConfig
from lmi.shell.LMIShellOptions import LMIShellOptions
from lmi.shell.LMISubscription import LMISubscription
from lmi.shell.LMINamespace import LMINamespace
from lmi.shell.LMINamespace import LMINamespaceRoot
from lmi.shell.LMIConstantValues import LMIConstantValues
from lmi.shell.LMIClass import LMIClass
from lmi.shell.LMIInstanceName import LMIInstanceName
from lmi.shell.LMIMethod import LMIMethod
from lmi.shell.LMIInstance import LMIInstance
from lmi.shell.LMIConnection import LMIConnection

from lmi.shell.LMIConnection import connect

# Register LMI wrapper classes into a LMIObjectFactory  to avoid circular
# dependency between those classes.
from lmi.shell.LMIObjectFactory import LMIObjectFactory
LMIObjectFactory().register(LMINamespace)
LMIObjectFactory().register(LMINamespaceRoot)
LMIObjectFactory().register(LMIClass)
LMIObjectFactory().register(LMIInstance)
LMIObjectFactory().register(LMIInstanceName)
LMIObjectFactory().register(LMIMethod)

# LMIConsole depends on LMIObjectFactory and needs to be imported after
# LMIObjectFactory has all wrapper classes registered.
from lmi.shell.LMIConsole import LMIConsole



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys

from lmi.shell.compat import *

from lmi.shell.LMIBaseObject import LMIWrapperBaseObject
from lmi.shell.LMIFormatter import LMIInstanceFormatter
from lmi.shell.LMIFormatter import LMIMofFormatter
from lmi.shell.LMIObjectFactory import LMIObjectFactory
from lmi.shell.LMIReturnValue import LMIReturnValue

from lmi.shell.LMIDecorators import lmi_possibly_deleted
from lmi.shell.LMIDecorators import lmi_return_expr_if_fail
from lmi.shell.LMIDecorators import lmi_return_val_if_fail
from lmi.shell.LMIDecorators import lmi_return_if_fail

from lmi.shell.LMIUtil import lmi_cast_to_cim
from lmi.shell.LMIUtil import lmi_cast_to_lmi
from lmi.shell.LMIUtil import lmi_transform_to_lmi
from lmi.shell.LMIUtil import lmi_wrap_cim_method


class LMIInstance(LMIWrapperBaseObject):
    """
    LMI wrapper class representing :py:class:`wbem.CIMInstance`.

    :param LMIConnection conn: connection object
    :param LMIClass lmi_class: wrapped creation class of the instance
    :param CIMInstance cim_instance: wrapped object
    """
    def __init__(self, conn, lmi_class, cim_instance):
        # We use __dict__ to avoid recursion potentially caused by
        # combo __setattr__ and __getattr__
        if isinstance(cim_instance, LMIInstance):
            cim_instance = cim_instance.wrapped_object
        self.__dict__["_deleted"] = False
        self.__dict__["_cim_instance"] = cim_instance
        self.__dict__["_lmi_class"] = lmi_class
        super(LMIInstance, self).__init__(conn)

    def __cmp__(self, other):
        """
        :param LMIInstance other: :py:class:`.LMIInstance` object to compare
        :returns: If both instances are (not) deleted: negative number, if
            self < other; 0 if self == other or positive number,
            if self > other. If the first or second instance is deleted, -1 or
            1 is returned.
        :rtype: int
        """
        if not isinstance(other, LMIInstance):
            return -1
        if self._deleted and not other._deleted:
            return -1
        elif not self._deleted and other._deleted:
            return 1
        return cmp(self._cim_instance, other._cim_instance)

    @lmi_possibly_deleted(False)
    def __contains__(self, key):
        """
        Returns True, if the specified key is present in the properties, False
        otherwise.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string key: key name, which will be tested for presence in
            properties
        :returns: True, if the specified key is present in properties, False
            otherwise
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return key in self._cim_instance

    @lmi_possibly_deleted(None)
    def __getattr__(self, name):
        """
        Returns either a class member, :py:class:`.LMIMethod` object, or a
        :py:class:`wbem.CIMInstance` object property.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string name: class member, the method name, or the property to
            be returned
        :returns: class member, :py:class:`.LMIMethod` object or
            :py:class:`wbem.CIMInstance` object property
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        if name in self._cim_instance:
            member = self._cim_instance.properties[name]
            if isinstance(member.value, wbem.CIMInstanceName):
                return lmi_transform_to_lmi(self._conn, member.value)
            return lmi_cast_to_lmi(member.type, member.value)
        elif not self._conn.is_wsman() and name in self.methods():
            return lmi_wrap_cim_method(self._conn, name, self)
        elif self._conn.is_wsman() and not name.startswith("_"):
            return lmi_wrap_cim_method(self._conn, name, self)
        raise AttributeError(name)

    @lmi_possibly_deleted(None)
    def __setattr__(self, name, value):
        """
        Modifies a :py:class:`wbem.CIMInstance` object property.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string name: name of the :py:class:`wbem.CIMInstance`'s member
        :param value: new value
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        if isinstance(value, LMIObjectFactory().LMIInstanceName):
            value = value.wrapped_object
        if name in self._cim_instance:
            t = self._cim_instance.properties[name].type
            self._cim_instance[name] = lmi_cast_to_cim(t, value)
        else:
            self.__dict__[name] = value

    @lmi_possibly_deleted("")
    def __str__(self):
        """
        Returns a string containing object path.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return an empty string. If
        the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.

        :returns: string containing object path
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return unicode(self).encode("utf-8")

    @lmi_possibly_deleted(u"")
    def __unicode__(self):
        """
        Returns an unicode string containing object path.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return an empty string. If
        the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.

        :returns: unicode string containing object path
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return unicode(self._cim_instance.path)

    @lmi_possibly_deleted("")
    def __repr__(self):
        """
        Returns a pretty string for the object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return an empty string. If
        the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.

        :returns: pretty string for the object
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return "%s(classname=\"%s\", ...)" % (
            self.__class__.__name__, self.classname)

    def copy(self):
        """
        :returns: copy of itself
        """
        cim_instance = self._cim_instance.copy()
        return lmi_transform_to_lmi(self._conn, cim_instance)

    @property
    @lmi_possibly_deleted("")
    def classname(self):
        """
        Property returning a string of a class name.

        **NOTE:** If the method :py:meth:`LMIInstance.delete` was called, this
        method will not execute its code and will return an empty string. If
        the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.

        :returns: class name
        :rtype: string
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return self._lmi_class.classname

    @property
    @lmi_possibly_deleted("")
    def namespace(self):
        """
        Property retuning a string of a namespace name.

        **NOTE:** If the method :py:meth:`LMIInstance.delete` was called, this
        method will not execute its code and will return an empty string. If
        the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.

        :returns: namespace name
        :rtype: string
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return self._lmi_class.namespace

    @property
    @lmi_possibly_deleted(None)
    def path(self):
        """
        Property returning a :py:class:`.LMIInstanceName` object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: :py:class:`.LMIInstanceName` object
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return lmi_transform_to_lmi(self._conn, self._cim_instance.path)

    @lmi_possibly_deleted(None)
    @lmi_return_if_fail(lambda obj: obj._cim_instance)
    def doc(self):
        """
        Prints out pretty verbose message with documentation for the instance.
        If the LMIShell is run in a interactive mode, the output will be
        redirected to a pager set by environment variable :envvar:`PAGER`. If
        there is not :envvar:`PAGER` set, less or more will be used as a
        fall-back.

        **NOTE:** If the method :py:meth:`LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        formatter = LMIInstanceFormatter(self._cim_instance)
        formatter.fancy_format(self._conn.client.interactive)

    @lmi_possibly_deleted(None)
    @lmi_return_if_fail(lambda obj: obj._cim_instance)
    def tomof(self):
        """
        Prints out a message with MOF representation of
        :py:class:`wbem.CIMMethod`.  If the LMIShell is run in a interactive
        mode, the output will be redirected to a pager set by environment
        variable :envvar:`PAGER`. If there is not :envvar:`PAGER` set, less or
        more will be used as a fall-back.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        formatter = LMIMofFormatter(self._cim_instance)
        formatter.fancy_format(self._conn.client.interactive)

    @lmi_possibly_deleted([])
    @lmi_return_val_if_fail(lambda obj: obj._cim_instance.path, [])
    def associator_names(self, AssocClass=None, ResultClass=None, Role=None,
                         ResultRole=None):
        """
        Returns a list of associated :py:class:`.LMIInstanceName` with this
        object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return an empty list. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string AssocClass: valid CIM association class name. It acts as
            a filter on the returned set of names by mandating that each
            returned name identify an object that shall be associated to the
            source object through an instance of this class or one of its
            subclasses.
        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be either an instance of this class
            (or one of its subclasses) or be this class (or one of its
            subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of names by mandating that each returned name identify
            an object that shall be associated to the source object through an
            association in which the source object plays the specified role.
            That is, the name of the property in the association class that
            refers to the source object shall match the value of this
            parameter.
        :param string ResultRole: valid property name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be associated to the source object
            through an association in which the named returned object plays the
            specified role. That is, the name of the property in the
            association class that refers to the returned object shall match
            the value of this parameter.
        :returns: list of associated :py:class:`.LMIInstanceName` objects
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`associators_instance_names`.
        """
        assoc_names = self._conn.client.get_associator_names(
            self._cim_instance, AssocClass, ResultClass, Role, ResultRole)
        return lmi_transform_to_lmi(self._conn, assoc_names)

    @lmi_possibly_deleted(None)
    def first_associator_name(self, AssocClass=None, ResultClass=None, Role=None,
                              ResultRole=None):
        """
        Returns the first associated :py:class:`.LMIInstanceName` with this
        object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string AssocClass: valid CIM association class name. It acts as
            a filter on the returned set of names by mandating that each
            returned name identify an object that shall be associated to the
            source object through an instance of this class or one of its
            subclasses.
        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be either an instance of this class
            (or one of its subclasses) or be this class (or one of its
            subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of names by mandating that each returned name identify
            an object that shall be associated to the source object through an
            association in which the source object plays the specified role.
            That is, the name of the property in the association class that
            refers to the source object shall match the value of this
            parameter.
        :param string ResultRole: valid property name. It acts as a filter on
            the returned set of names by mandating that each returned name
            identify an object that shall be associated to the source object
            through an association in which the named returned object plays the
            specified role. That is, the name of the property in the
            association class that refers to the returned object shall match
            the value of this parameter.
        :returns: first associated :py:class:`.LMIInstanceName` object
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`associators_instance_names`.
        """
        assoc_names = self._conn.client.get_associator_names(
            self._cim_instance, AssocClass, ResultClass, Role, ResultRole,
            limit=1)
        if not assoc_names:
            return None
        return lmi_transform_to_lmi(self._conn, assoc_names[0])

    @lmi_possibly_deleted([])
    @lmi_return_val_if_fail(lambda obj: obj._cim_instance.path, [])
    def associators(self, AssocClass=None, ResultClass=None, Role=None,
                    ResultRole=None, IncludeQualifiers=False,
                    IncludeClassOrigin=False, PropertyList=None):
        """
        Returns a list of associated :py:class:`.LMIInstance` objects with this
        instance.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return an empty list. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string AssocClass: valid CIM association class name. It acts as
            a filter on the returned set of objects by mandating that each
            returned object shall be associated to the source object through an
            instance of this class or one of its subclasses.
        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be either an instance of this class (or one of its
            subclasses) or be this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of objects by mandating that each returned object
            shall be associated with the source object through an association
            in which the source object plays the specified role. That is, the
            name of the property in the association class that refers to the
            source object shall match the value of this parameter.
        :param string ResultRole: valid property name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be associated to the source object through an association in
            which the returned object plays the specified role. That is, the
            name of the property in the association class that refers to the
            returned object shall match the value of this parameter.
        :param bool IncludeQualifiers: bool flag indicating, if all qualifiers
            for each object (including qualifiers on the object and on any
            returned properties) shall be included as ``<QUALIFIER>`` elements
            in the response.
        :param bool IncludeClassOrigin: bool flag indicating, if the
            ``CLASSORIGIN`` attribute shall be present on all appropriate
            elements in each returned object.
        :param list PropertyList: if not None, the members of the array define
            one or more property names. Each returned object shall not include
            elements for any properties missing from this list. If
            *PropertyList* is an empty list, no properties are included in each
            returned object.  If it is None, no additional filtering is
            defined.
        :returns: list of associated :py:class:`.LMIInstance` objects
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`associators_instances`.
        """
        associators = self._conn.client.get_associators(
            self._cim_instance,
            AssocClass, ResultClass, Role, ResultRole, IncludeQualifiers,
            IncludeClassOrigin, PropertyList)
        return lmi_transform_to_lmi(self._conn, associators)

    @lmi_possibly_deleted(None)
    def first_associator(self, AssocClass=None, ResultClass=None, Role=None,
                         ResultRole=None, IncludeQualifiers=False,
                         IncludeClassOrigin=False, PropertyList=None):
        """
        Returns the first associated :py:class:`.LMIInstance` with this object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string AssocClass: valid CIM association class name. It acts as
            a filter on the returned set of objects by mandating that each
            returned object shall be associated to the source object through an
            instance of this class or one of its subclasses.
        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be either an instance of this class (or one of its
            subclasses) or be this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of objects by mandating that each returned object
            shall be associated with the source object through an association
            in which the source object plays the specified role. That is, the
            name of the property in the association class that refers to the
            source object shall match the value of this parameter.
        :param string ResultRole: valid property name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be associated to the source object through an association in
            which the returned object plays the specified role. That is, the
            name of the property in the association class that refers to the
            returned object shall match the value of this parameter.
        :param bool IncludeQualifiers: bool flag indicating, if all qualifiers
            for each object (including qualifiers on the object and on any
            returned properties) shall be included as ``<QUALIFIER>`` elements
            in the response.
        :param bool IncludeClassOrigin: bool flag indicating, if the
            ``CLASSORIGIN`` attribute shall be present on all appropriate
            elements in each returned object.
        :param list PropertyList: if not None, the members of the array define
            one or more property names. Each returned object shall not include
            elements for any properties missing from this list. If PropertyList
            is an empty list, no properties are included in each returned
            object. If it is None, no additional filtering is defined.
        :returns: first associated :py:class:`.LMIInstance`
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`associators_instances`.
        """
        associators = self._conn.client.get_associators(
            self._cim_instance,
            AssocClass, ResultClass, Role, ResultRole, IncludeQualifiers,
            IncludeClassOrigin, PropertyList, limit=1)
        if not associators:
            return None
        return lmi_transform_to_lmi(self._conn, associators[0])

    @lmi_possibly_deleted([])
    @lmi_return_val_if_fail(lambda obj: obj._cim_instance.path, [])
    def reference_names(self, ResultClass=None, Role=None):
        """
        Returns a list of association :py:class:`.LMIInstanceName` objects with
        this object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return an empty list. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of object names by mandating that each returned
            Object Name identify an instance of this class (or one of its
            subclasses) or this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of object names by mandating that each returned object
            name shall identify an object that refers to the target instance
            through a property with a name that matches the value of this
            parameter.
        :returns: list of association :py:class:`.LMIInstanceName` objects
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`references_instance_names`.
        """
        reference_names = self._conn.client.get_reference_names(
            self._cim_instance, ResultClass, Role)
        return lmi_transform_to_lmi(self._conn, reference_names)

    @lmi_possibly_deleted(None)
    def first_reference_name(self, ResultClass=None, Role=None):
        """
        Returns the first association :py:class:`.LMIInstanceName` with this
        object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of object names by mandating that each returned
            Object Name identify an instance of this class (or one of its
            subclasses) or this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of object names by mandating that each returned object
            name shall identify an object that refers to the target instance
            through a property with a name that matches the value of this
            parameter.
        :returns: first association :py:class:`.LMIInstanceName` object
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`references_instance_names`.
        """
        reference_names = self._conn.client.get_reference_names(
            self._cim_instance, ResultClass, Role, limit=1)
        if not reference_names:
            return None
        return lmi_transform_to_lmi(self._conn, reference_names[0])

    @lmi_possibly_deleted([])
    @lmi_return_val_if_fail(lambda obj: obj._cim_instance.path, [])
    def references(self, ResultClass=None, Role=None, IncludeQualifiers=False,
                   IncludeClassOrigin=False, PropertyList=None):
        """
        Returns a list of association :py:class:`.LMIInstance` objects with this
        object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return an empty list. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be an instance of this class (or one of its subclasses) or
            this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of objects by mandating that each returned object
            shall refer to the target object through a property with a name
            that matches the value of this parameter.
        :param bool IncludeQualifiers: flag indicating, if all qualifiers for
            each object (including qualifiers on the object and on any returned
            properties) shall be included as ``<QUALIFIER>`` elements in the
            response.
        :param bool IncludeClassOrigin: flag indicating, if the ``CLASSORIGIN``
            attribute shall be present on all appropriate elements in each
            returned object.
        :param list PropertyList: if not None, the members of the list define
            one or more property names. Each returned object shall not include
            elements for any properties missing from this list. If PropertyList
            is an empty list, no properties are included in each returned
            object. If PropertyList is None, no additional filtering is
            defined.
        :returns: list of association :py:class:`.LMIInstance` objects
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`references_instances`.
        """
        references = self._conn.client.get_references(
            self._cim_instance, ResultClass, Role, IncludeQualifiers,
            IncludeClassOrigin, PropertyList)
        return lmi_transform_to_lmi(self._conn, references)

    @lmi_possibly_deleted(None)
    def first_reference(self, ResultClass=None, Role=None,
                        IncludeQualifiers=False, IncludeClassOrigin=False,
                        PropertyList=None):
        """
        Returns the first association :py:class:`.LMIInstance` with this object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string ResultClass: valid CIM class name. It acts as a filter on
            the returned set of objects by mandating that each returned object
            shall be an instance of this class (or one of its subclasses) or
            this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the
            returned set of objects by mandating that each returned object
            shall refer to the target object through a property with a name
            that matches the value of this parameter.
        :param bool IncludeQualifiers: flag indicating, if all qualifiers for
            each object (including qualifiers on the object and on any returned
            properties) shall be included as ``<QUALIFIER>`` elements in the
            response.
        :param bool IncludeClassOrigin: flag indicating, if the ``CLASSORIGIN``
            attribute shall be present on all appropriate elements in each
            returned object.
        :param list PropertyList: if not None, the members of the list define
            one or more property names. Each returned object shall not include
            elements for any properties missing from this list. If PropertyList
            is an empty list, no properties are included in each returned
            object. If PropertyList is None, no additional filtering is
            defined.
        :returns: first association :py:class:`.LMIInstance` object
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`references_instances`.
        """
        references = self._conn.client.get_references(
            self._cim_instance, ResultClass, Role, IncludeQualifiers,
            IncludeClassOrigin, PropertyList, limit=1)
        if not references:
            return None
        return lmi_transform_to_lmi(self._conn, references[0])

    @lmi_possibly_deleted([])
    @lmi_return_val_if_fail(lambda obj: obj._cim_instance.properties, [])
    def properties(self):
        """
        Returns a list of :py:class:`wbem.CIMInstance` properties.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return an empty list. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: list of :py:class:`wbem.CIMInstance` properties
        :rtype: list
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_properties`.
        """
        return self._cim_instance.properties.keys()

    @lmi_possibly_deleted(None)
    @lmi_return_if_fail(lambda obj: obj._cim_instance.properties)
    def print_properties(self):
        """
        Prints out the list of :py:class:`wbem.CIMInstance` properties.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_properties`.
        """
        for name, prop in self._cim_instance.properties.iteritems():
            value = str(prop.value)
            if isinstance(prop.value, basestring):
                value = "'%s'" % value
            sys.stdout.write("%s = %s\n" % (name, value))

    @lmi_possibly_deleted({})
    @lmi_return_val_if_fail(lambda obj: obj._cim_instance.properties, {})
    def properties_dict(self):
        """
        Returns dictionary containing property name and value pairs.
        This method may consume significant memory amount when called.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return an empty dictionary.
        If the shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be
        raised.

        :returns: dictionary of :py:class:`wbem.CIMInstance` properties
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        props = self._cim_instance.properties
        return wbem.NocaseDict(
            dict((k, x.value) for k, x in props.iteritems()))

    @lmi_possibly_deleted(None)
    def property_value(self, prop_name):
        """
        Returns a :py:class:`wbem.CIMInstance` property value.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string prop_name: :py:class:`wbem.CIMInstance` property name
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return getattr(self, prop_name)

    @lmi_possibly_deleted([])
    def methods(self):
        """
        Returns a list of :py:class:`wbem.CIMInstance` methods' names.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return an empty list. If the
        shell uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: list of :py:class:`wbem.CIMInstance` methods' names
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_methods`.
        """
        return self._lmi_class.methods()

    @lmi_possibly_deleted(None)
    def print_methods(self):
        """
        Prints out the list of :py:class:`wbem.CIMInstance` methods' names.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_methods`.
        """
        self._lmi_class.print_methods()

    @lmi_possibly_deleted(
        LMIReturnValue(
            rval=False,
            errorstr="This instance has been deleted from a CIM broker"))
    @lmi_return_val_if_fail(
        lambda obj: obj._cim_instance,
        LMIReturnValue(rval=False, errorstr="Can not refresh the instance"))
    def refresh(self):
        """
        Retrieves a new :py:class:`wbem.CIMInstance` object. Basically
        refreshes the object properties. Returns :py:class:`.LMIReturnValue`
        with rval set to 0, if the wrapped :py:class:`wbem.CIMInstance` object
        was refreshed; otherwise rval is set to -1.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return
        :py:class:`.LMIReturnValue` object containing -1 as a return value with
        proper error string set. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to 0, if
            refreshed; -1 otherwise
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_refreshing`.
        """
        new_cim_instance, _, errorstr = self._conn.client.get_instance(
            self.path, LocalOnly=False)
        if not new_cim_instance:
            return LMIReturnValue(rval=False, errorstr=errorstr)
        self._cim_instance = new_cim_instance
        return LMIReturnValue(rval=True)

    @lmi_possibly_deleted(
        LMIReturnValue(
            rval=False,
            errorstr="This instance has been deleted from a CIM broker"))
    @lmi_return_val_if_fail(
        lambda obj: obj._cim_instance.path,
        LMIReturnValue(rval=False, errorstr="Can't push this instance"))
    def push(self):
        """
        Pushes the modified object to the CIMOM.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return
        :py:class:`.LMIReturnValue` object containing False as a return value
        with proper error string set. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: :py:class:`.LMIReturnValue` object with ``rval`` set to True, if
            modified; False otherwise
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_properties`.
        """
        return self._conn.client.modify_instance(self._cim_instance)

    @lmi_possibly_deleted(None)
    @lmi_return_if_fail(lambda obj: obj._cim_instance.path)
    def delete(self):
        """
        Deletes this instance from the CIMOM.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this
        method will not execute its code and will return None. If the shell
        uses exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: True, if the instance is deleted; False otherwise
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_delete`.
        """
        rval, rparams, errorstr = self._conn.client.delete_instance(
            self._cim_instance.path)
        self._deleted = rval
        return self._deleted

    @property
    def is_deleted(self):
        """
        :returns: True, if the instance was deleted from the CIMOM; False
            otherwise
        """
        return self._deleted

    @property
    @lmi_possibly_deleted(None)
    def wrapped_object(self):
        """
        :returns: wrapped :py:class:`wbem.CIMInstance` object
        """
        return self._cim_instance



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Module for session object representing all connection to remote hosts.
"""

from collections import defaultdict

from lmi.scripts.common import errors
from lmi.scripts.common import get_logger
from lmi.scripts.common.util import FilteredDict
from lmi.shell.LMIConnection import connect

LOG = get_logger(__name__)

class Session(object):
    """
    Session object keeps connection objects to remote hosts. Their are
    associated with particular hostnames. It also caches credentials for them.
    Connections are made as they are needed. When credentials are missing
    for connection to be made, the user is asked to supply them from
    standard input.

    :param app: Instance of main application.
    :param list hosts: List of hostname strings.
    :param dictionary credentials: Mapping assigning a pair
        ``(user, password)`` to each hostname.
    :param boolean same_credentials: Use the same credentials for all
        hosts in session. The first credentials given will be used.
    """

    def __init__(self, app, hosts, credentials=None, same_credentials=False):
        self._app = app
        self._connections = dict((h, None) for h in hosts)
        # { hostname : (username, password, verified), ... }
        # where verified is a flag saying, whether these credentials
        # were successfuly used for logging in
        self._credentials = defaultdict(lambda: ('', '', False))
        self._same_credentials = same_credentials
        if credentials is not None:
            if not isinstance(credentials, dict):
                raise TypeError("credentials must be a dictionary")
            for hostname, creds in credentials.items():
                credentials[hostname] = (creds[0], creds[1], False)
            self._credentials.update(credentials)

    def __getitem__(self, hostname):
        """
        :rtype: (``LMIConnection``) Connection object to remote host.
            ``None`` if connection can not be made.
        """
        if self._connections[hostname] is None:
            try:
                self._connections[hostname] = self._connect(
                        hostname, interactive=True)
            except Exception as exc:
                LOG().error('Failed to make a connection to "%s": %s',
                        hostname, exc)
        return self._connections[hostname]

    def __len__(self):
        """ Get the number of hostnames in session. """
        return len(self._connections)

    def __contains__(self, uri):
        return uri in self._connections

    def __iter__(self):
        """ Yields connection objects. """
        successful_connections = 0
        for hostname in self._connections:
            connection = self[hostname]
            if connection is not None:
                yield connection
                successful_connections += 1
        if successful_connections == 0:
            raise errors.LmiNoConnections('No successful connection made.')

    def _connect(self, hostname, interactive=False):
        """
        Makes the connection to host.

        :param string hostname: Name of host.
        :param boolean interactive: Whether we can interact with user
            and expect a reply from him.
        :returns: Connection to remote host or ``None``.
        :rtype: :py:class:`lmi.shell.LMIConnection` or ``None``
        """
        username, password = self.get_credentials(hostname)
        kwargs = {
                'verify_server_cert' : self._app.config.verify_server_cert,
                'interactive'        : interactive
        }
        if len(self._connections) > 1:
            kwargs['prompt_prefix'] = '[%s] ' % hostname
        connection = connect(hostname, username, password, **kwargs)
        if connection is not None:
            tp = connection.client._cliconn.creds
            if tp is None:
                tp = ('', '')
            self._credentials[hostname] = (tp[0], tp[1], True)
        else:
            LOG().error('Failed to connect to host "%s".', hostname)
        return connection

    @property
    def hostnames(self):
        """
        List of hostnames in session.

        :rtype: list
        """
        return self._connections.keys()

    def get_credentials(self, hostname):
        """
        :param string hostname: Name of host to get credentials for.
        :returns: Pair of ``(username, password)`` for given hostname. If no
            credentials were given for this host, ``('', '')`` is returned.
        :rtype: tuple
        """
        username, password, verified = self._credentials[hostname]
        if (   not verified
           and (not username or not password)
           and self._same_credentials):
            for tp in self._credentials.values():
                if tp[2]:
                    username, password = tp[0], tp[1]
                    break
        return username, password

    def get_unconnected(self):
        """
        :returns:  List of hostnames, which do not have associated connection
            yet.
        :rtype: list
        """
        return [h for h, c in self._connections.items() if c is None]

class SessionProxy(Session):
    """
    Behaves like a session. But it just encapsulates other session object and
    provides access to a subset of its items.

    :param session: Session object or even another session proxy.
    :param list uris: Subset of uris in encapsulated session object.
    """

    def __init__(self, session, uris):
        uris = set(uris)
        if not all(isinstance(uri, basestring) for uri in uris):
            raise ValueError("uris must be iterable of uris")
        for uri in uris:
            if not uri in session:
                raise ValueError('uri "%s" needs to belong to given session'
                        % uri)
        Session.__init__(self, session._app, uris, session._credentials,
                session._same_credentials)
        self._origin = session
        self._connections = FilteredDict(uris, session._connections)
        # let the credentials propagage to original session
        self._credentials = session._credentials




# Copyright (C) 2013-2014 Red Hat, Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of the FreeBSD Project.
#
# Authors: Michal Minar <miminar@redhat.com>
#
"""
Various utilities for LMI Scripts.
"""

class FilteredDict(dict):
    """
    Dictionary-like collection that wraps some other dictionary and provides
    limited access to its keys and values. It permits to get, delete and set
    items specified in advance.

    .. note::
        Please use only the methods overriden. This class does not guarantee
        100% API compliance. Not overriden methods won't work properly.

    :param list key_filter: Set of keys that can be get, set or deleted.
        For other keys, :py:class:`KeyError` will be raised.
    :param dictionary original: Original dictionary containing not only
        keys in *key_filter* but others as well. All modifying operations
        operate also on this dictionary. But only those keys in *key_filter*
        can be affected by them.
    """

    def __init__(self, key_filter, original=None):
        dict.__init__(self)
        if original is not None and not isinstance(original, dict):
            raise TypeError("original needs to be a dictionary")
        if original is None:
            original = dict()
        self._original = original
        self._keys = frozenset(key_filter)

    def __contains__(self, key):
        return key in self._keys and key in self._original

    def __delitem__(self, key):
        if not key in self._keys:
            raise KeyError(repr(key))
        del self._original[key]

    def clear(self):
        for key in self._keys:
            self._original.pop(key, None)

    def copy(self):
        return FilteredDict(self._keys, self._original)

    def iterkeys(self):
        for key in self._keys:
            yield key

    def __getitem__(self, key):
        if not key in self._keys:
            raise KeyError(repr(key))
        return self._original[key]

    def __iter__(self):
        for key in self._keys:
            if key in self._original:
                yield key

    def __eq__(self, other):
        return (   isinstance(other, FilteredDict)
               and self._original == other._original
               and self._keys == other._keys)

    def __lt__(self, other):
        if isinstance(other, dict):
            return dict((k, v) for k, v in self._original.items()
                   if k in self._keys) < other
        if not isinstance(other, FilteredDict):
            raise TypeError("Can not compare FilteredDict to objects"
                    " of other types!")
        return self._original <= other._original and self._keys <= other._keys

    def __len__(self):
        return len(self.keys())

    def __setitem__(self, key, value):
        if not key in self._keys:
            raise KeyError(repr(key))
        self._original[key] = value

    def keys(self):
        return [k for k in self._keys if k in self._original]

    def values(self):
        return [self._original[k] for k in self.keys() if k in self._original]

    def items(self):
        return [(k, self._original[k]) for k in self.keys()]

    def iteritems(self):
        return iter(self.items())

    def pop(self, key, *args):
        ret = self[key]
        del self[key]
        return ret

    def popitem(self):
        for key in self._keys:
            if key in self._original:
                return self.pop(key)
        raise KeyError("FilterDict is empty!")

    def update(self, *args, **kwargs):
        if len(args) > 1:
            raise TypeError('Expected just one positional argument!')
        if args and callable(getattr(args[0], 'keys', None)):
            for key in args[0].keys():
                self[key] = args[0][key]
        elif args:
            for key, value in args[0]:
                self[key] = value
        for key, value in kwargs.items():
            self[key] = value




# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Utilities for logging framework.
"""
import logging
import sys

from lmi.scripts.common import configuration
from lmi.scripts.common import errors

TERM_COLOR_NAMES = (
        'black',
        'red',
        'green',
        'yellow',
        'blue',
        'magenta',
        'cyan',
        'white',
        )

# Following are terminal color codes of normal mode.
CN_BLACK   = 0
CN_RED     = 1
CN_GREEN   = 2
CN_YELLOW  = 3
CN_BLUE    = 4
CN_MAGENTA = 5
CN_CYAN    = 6
CN_WHITE   = 7
# Following are terminal color codes of bright mode.
CB_BLACK   = 8
CB_RED     = 9
CB_GREEN   = 10
CB_YELLOW  = 11
CB_BLUE    = 12
CB_MAGENTA = 13
CB_CYAN    = 14
CB_WHITE   = 15

WARNING_COLOR  = CB_YELLOW
ERROR_COLOR    = CB_RED
CRITICAL_COLOR = CB_MAGENTA

#: Dictionary assigning color code to logging level.
LOG_LEVEL_2_COLOR = {
        logging.WARNING : WARNING_COLOR,
        logging.ERROR   : ERROR_COLOR,
        logging.CRITICAL: CRITICAL_COLOR
}

class LogRecord(logging.LogRecord):
    """
    Overrides :py:class:`logging.LogRecord`. It adds new attributes:

        * `levelname_` - Name of level in lowercase.
        * `cseq`       - Escape sequence for terminal used to set color
                         assigned to particular log level.
        * `creset`     - Escape sequence for terminal used to reset foreground
                         color.

    These can be used in format strings initializing logging formatters.

    Accepts the same arguments as base class.
    """

    def __init__(self, name, level, *args, **kwargs):
        use_colors = kwargs.pop('use_colors', True)
        logging.LogRecord.__init__(self, name, level, *args, **kwargs)
        self.levelname_ = self.levelname.lower()
        if use_colors and level >= logging.WARNING:
            if level >= logging.CRITICAL:
                color = LOG_LEVEL_2_COLOR[logging.CRITICAL]
            elif level >= logging.ERROR:
                color = LOG_LEVEL_2_COLOR[logging.ERROR]
            else:
                color = LOG_LEVEL_2_COLOR[logging.WARNING]
            self.cseq = get_color_sequence(color)
            self.creset = "\x1b[39m"
        else:
            self.cseq = self.creset = ''

class ScriptsLogger(logging.getLoggerClass()):

    #: Boolean saying whether the color sequences for log messages shall be
    #: generated.
    USE_COLORS = True

    def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None,
            extra=None):
        """
        Overriden method that just changes the *LogRecord* class to our
        predefined and ensures that exception's traceback is printed once at
        most.
        """
        if exc_info:
            err = exc_info[1]
            if getattr(err, '_traceback_logged', False):
                # do not print traceback second time
                exc_info = None
            elif self.isEnabledFor(level):
                try:
                    err._traceback_logged = True
                except AttributeError:
                    pass
        rv = LogRecord(name, level, fn, lno, msg, args, exc_info, func,
                use_colors=self.USE_COLORS)
        if extra is not None:
            for key in extra:
                if (key in ["message", "asctime"]) or (key in rv.__dict__):
                    raise KeyError("Attempt to overwrite %r in LogRecord" % key)
                rv.__dict__[key] = extra[key]
        return rv

    def exception(self, msg, *args, **kwargs):
        lmi_config = configuration.Configuration.get_instance()
        exc_info = sys.exc_info()
        err = exc_info[1]
        if (   lmi_config.trace
           and (  not isinstance(err, errors.LmiError)
               or (   not getattr(err, '_traceback_logged', False)
                  and lmi_config.verbosity >= lmi_config.OUTPUT_DEBUG))):
            kwargs['exc_info'] = exc_info
        else:
            kwargs.pop('exc_info', None)
        self.error(msg, *args, **kwargs)

class LevelDispatchingFormatter(object):
    """
    Formatter class for logging module. It allows to predefine different format
    string used for some level ranges.

    :param dict formatters: Mapping of module names to *format*.
        It is a dictionary of following format: ::

            { max_level1 : format1
            , max_level2 : format2
            , ... }

        *format* in parameters description can be either ``string`` or another
        formatter object.

        For example if you want to have format3 used for *ERROR* and *CRITICAL*
        levels, *format2* for *INFO* and *format1* for anything else, your
        dictionary will look like this: ::

            { logging.INFO - 1 : format1
            , logging.INFO     : format2 }

        And the *default* value should have *format3* assigned.

    :param default: Default *format* to use. This format is used for all levels
        higher than the maximum of *formatters*' keys.
    """
    def __init__(self, formatters, default=configuration.DEFAULT_FORMAT_STRING,
            datefmt=None):
        for k, formatter in formatters.items():
            if isinstance(formatter, basestring):
                formatters[k] = logging.Formatter(formatter, datefmt=datefmt)
        #: This a a tuple of pairs sorted by the first item in descending
        #: order (highest priority ordered first).
        self._formatters = tuple(sorted(formatters.items(),
            key=lambda t: -t[0]))
        if isinstance(default, basestring):
            default = logging.Formatter(default, datefmt=datefmt)
        self._default_formatter = default

    def format(self, record):
        """
        Interface for logging module.
        """
        record.levelname_ = record.levelname.lower()
        formatter = self._default_formatter
        for level, fmt in self._formatters:
            if level < record.levelno:
                break
            formatter = fmt
        try:
            return formatter.format(record)
        except KeyError:
            # in some modules or libraries it may happen that logger is
            # initialized before our logger class is set as default
            record.cseq = record.creset = ''
            return formatter.format(record)

def get_logger(module_name):
    """
    Convenience function for getting callable returning logger for particular
    module name. It's supposed to be used at module's level to assign its
    result to global variable like this: ::

        from lmi.scripts import common

        LOG = common.get_logger(__name__)

    This can be used in module's functions and classes like this: ::

        def module_function(param):
            LOG().debug("This is debug statement logging param: %s", param)

    Thanks to ``LOG`` being a callable, it always returns valid logger object
    with current configuration, which may change overtime.

    :param string module_name: Absolute dotted module path.
    :rtype: :py:class:`logging.Logger`
    """
    def _logger():
        """ Callable used to obtain current logger object. """
        return logging.getLogger(module_name)
    return _logger

def get_color_sequence(color_code):
    """
    Computer color sequence for particular color code.

    :returns: Escape sequence for terminal used to set foreground color.
    :rtype: str
    """
    if color_code <= 7:
        return "\x1b[%dm" % (30 + color_code)
    return "\x1b[38;5;%dm" % color_code

def setup_logger(use_colors=True):
    """
    This needs to be called before any logging takes place.
    """
    ScriptsLogger.USE_COLORS = use_colors
    logging.setLoggerClass(ScriptsLogger)



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
LMIShell compatibility module. It imports WBEM backend modules and sets
necessary attributes to either LMIWBEM or PyWBEM.
"""

import sys

from lmi.shell.LMIShellLogger import lmi_get_logger

logger = lmi_get_logger()

# First, try to import LMIWBEM module.
try:
    HAVE_LMIWBEM = True

    import lmiwbem as wbem

    logger.debug("Using LMIWBEM backend")

    class Error(Exception):
        pass

    class AuthError(Error):
        pass

    # We add thse exceptions to LMIWBEM module, so that exception handlers in
    # lmi.shell.LMIDecorators can catch also these types. They exist due to
    # PyWBEM module.
    setattr(wbem, "Error", Error)
    setattr(wbem, "AuthError", AuthError)
except ImportError, e:
    HAVE_LMIWBEM = False

    # LMIWBEM could not be imported, fall back to PyWBEM.
    try:
        import pywbem as wbem
    except ImportError, e:
        # We can't import any backend, exit with error!
        logger.error("Can't import either lmiwbem or pywbem module.")
        sys.exit(1)

    logger.debug("Using PyWBEM backend")

    from lmi.shell.compat.CIMIndicationListener import CIMIndicationListener

    class ConnectionError(Exception):
        pass

    class NocaseDict(wbem.NocaseDict):
        """
        :py:class:`pywbem.NocaseDict` with compatibility API
        """
        def get(self, key, default=None):
            if key in self:
                return self[key]
            else:
                return default

        def pop(self, key, default=None):
            if key in self:
                rval = self[key]
                del self[key]
                return rval
            else:
                return default

    # We add these exception to PyWBEM module, so that exception handlers in
    # lmi.shell.LMIDecorators can catch also these types. They exist due to
    # LMIWBEM module.
    setattr(wbem, "ConnectionError", ConnectionError)
    setattr(wbem, "AuthError", wbem.cim_http.AuthError)

    # PyWBEM does not define any indication listener. We need to define one in
    # this compat module and provide it to PyWBEM module.
    setattr(wbem, "CIMIndicationListener", CIMIndicationListener)

    # NocaseDict defined in PyWBEM does not provide .pop() and .get() methods.
    # So we define them in our subclass of pywbem.NocaseDict and add necessary
    # interface.
    delattr(wbem, "NocaseDict")
    setattr(wbem, "NocaseDict", NocaseDict)



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys
import ssl
import time
import pywbem
import random
import socket
import string
import threading

from SocketServer import BaseRequestHandler
from SocketServer import ThreadingMixIn
from BaseHTTPServer import HTTPServer
from BaseHTTPServer import BaseHTTPRequestHandler

from lmi.shell.LMIExceptions import ConnectionError


class CIMIndicationHandlerCallback(object):
    """
    Helper class, which stores indication handler callback with its arguments
    and keyword arguments.

    :param callback: callback, which is called, when a indication arrives
    :param tuple args: positional arguments for callback
    :param kwargs: keyword arguments for callback
    """
    def __init__(self, callback, *args, **kwargs):
        self.callback = callback
        self.args = args
        self.kwargs = kwargs


class CIMIndicationHandler(ThreadingMixIn, BaseHTTPRequestHandler):
    """
    Class representing indication handler. The class is derived from
    :py:class:`BaseHTTPRequestHandler`, because the indications are transported
    by http protocol, and from :py:class:`ThreadingMixIn`; the indication
    listener is designed as concurrent server to properly handle each incoming
    indication.
    """
    CIMLISTENER_PREFIX = "/CIMListener"

    def do_POST(self):
        """
        Overridden method, which is called, when a indication is received. It
        parses the indication message and searches for a matching handler for
        the indication name.  Each subscribed indication should have it's
        Destination set something similar:

        ``<schema>://<destination>/<indication_name>``

        where the ``indication_name`` is a string, which properly defines the
        indication.
        """
        # If HTTP/1.1 used, try to get a message body length. If Content-Length
        # header is not present, set length to -1; rfile.read() will block,
        # until the connection is closed. Knowing the exact content length
        # makes the routine to finish early, even if the TCP connection is not
        # closed.
        length = int(self.headers.get("Content-Length", "-1"))
        msg = self.rfile.read(length)
        tt = pywbem.parse_cim(pywbem.xml_to_tupletree(msg))
        message = tt[2]
        export_methods = {}
        if message[0].upper() != "MESSAGE":
            return

        message_params = message[2]
        if not message_params:
            return

        for param in message_params:
            if param[0].upper() != "SIMPLEEXPREQ":
                continue
            export_method_call = param[2]
            export_method_name = export_method_call[1]["NAME"]
            exp_params = export_method_call[2]
            export_method_params = {}
            for method_param in exp_params:
                export_method_params[method_param[0]] = method_param[1]
            export_methods[export_method_name] = export_method_params

        ind = None
        if export_methods:
            ind_dict = export_methods.values()[0]
            if "NewIndication" in ind_dict:
                ind = ind_dict["NewIndication"]

        path = self.path
        cimlistener_ind = path.find(self.CIMLISTENER_PREFIX)
        if cimlistener_ind != -1:
            # We got an indication with lmiwbem's CIMListener substring
            # present. We are using pywbem now, so it's necessary to strip the
            # prefix also with the substring (from the beginning). We try to
            # mimic Pegasus::CIMListener behavior here.
            path = path[cimlistener_ind + len(self.CIMLISTENER_PREFIX):]

        if path[0] == '/':
            # Path of the received indication may be prefixed by additional
            # slash. Drop it then.
            path = path[1:]

        if path in self.server._handlers:
            # We found the indication handler. Execute the callback.
            cb = self.server._handlers[path]
            cb.callback(ind, *cb.args, **cb.kwargs)


class CIMIndicationServer(ThreadingMixIn, HTTPServer):
    """
    Class representing indication server, derived from HTTPServer and designed
    as concurrent server.
    """
    daemon_threads = True


class CIMIndicationListener(object):
    """
    Class representing indication listener, which provides a unified API for
    the server startup and shutdown and for registering an indication handler.

    :param str hostname: bind hostname
    :param int port: listening port
    :param str cert_file: path to X509 certificate
    :param str key_file:  path to X509 private key; may be None, if
        cert_file also contains private key
    :param str trust_store: path to trust store
    """
    def __init__(self, hostname, port, certfile=None, keyfile=None,
                 trust_store=None):
        self._handlers = {}
        self._hostname = hostname
        self._port = port
        self._certfile = certfile
        self._keyfile = keyfile
        self._trust_store = trust_store

        self._server = None
        self._server_thread = None

    def start(self, retries=1):
        """
        Starts a indication listener.

        The indication listener runs in a newly-created thread.

        :param int retries: number of bind retries.
        """
        # Try to create a listener for retries-times.
        for i in xrange(retries):
            try:
                self._server = CIMIndicationServer(
                    (self._hostname, self._port), CIMIndicationHandler)
                break
            except socket.error as e:
                # We raise the exception, when we run out of retries or TCP port
                # is out of range.
                self._port += 1
                if i == retries - 1 or self._port >= 65536:
                    raise ConnectionError(e)
        self._server._handlers = self._handlers
        self._server_thread = threading.Thread(
            target=self._server.serve_forever)
        self._server_thread.daemon = True
        self._server_uses_ssl = False
        if self._certfile:
            self._server.socket = ssl.wrap_socket(
                self._server.socket,
                certfile=self._certfile,
                keyfile=self._keyfile,
                server_side=True,
                ca_certs=self._trust_store)
            self._server_uses_ssl = True
        self._server_thread.start()

    def stop(self):
        """
        Stops the indication listener.

        This method will also terminate the listener thread.
        """
        if self._server:
            self._server.shutdown()
        if self._server_thread:
            self._server_thread.join()
        self._server = None
        self._server_thread = None

    @property
    def is_alive(self):
        """
        :returns: flag indicating, if the indication listener is running
        :rtype: bool
        """
        if self._server_thread:
            return self._server_thread.is_alive()
        return False

    @property
    def uses_ssl(self):
        """
        :returns: True, if :py:class:`.CIMIndicationListener` uses secure
            connection; False otherwise
        :rtype: bool
        """
        if not self._server:
            return False
        return self._server_uses_ssl

    @property
    def hostname(self):
        """
        :returns: hostname or address, where the indication listener is waiting
            for messages
        :rtype: string
        """
        return self._hostname

    @property
    def port(self):
        """
        :returns: port, where the indication listener is waiting for messages
        :rtype: int
        """
        return self._port

    @property
    def handlers(self):
        """
        :returns: list of string of handler keys
        :rtype: list
        """
        return self._handlers.keys()

    def add_handler(self, handler_name, handler, *args, **kwargs):
        """
        Registers a handler into the indication listener. Returns a string,
        which is used for the indication recognition, when a message arrives.

        :param string handler_name: handler name
        :param handler: callable, which will be executed, when a indication is
            received
        :param tuple args: positional arguments for the handler
        :param dictionary kwargs: keyword arguments for the handler
        """
        self._handlers[handler_name] = CIMIndicationHandlerCallback(
            handler, *args, **kwargs)

    def remove_handler(self, name):
        """
        Removes a specified handler from the indication listener database.

        :param string name: indication name; returned by
            :py:meth:`.CIMIndicationListener.add_handler`
        """
        if name not in self._handlers:
            raise KeyError("No such handler registered: %s" % name)
        self._handlers.pop(name)



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.


class LMIDeletedObjectError(Exception):
    """
    Raised, when there is an attempt to access properties on deleted
    :py:class:`.LMIInstance` object.
    """


class LMIUnknownParameterError(Exception):
    """
    Raised, when there is a method call issued and unknown method parameter is
    provided.
    """


class LMIUnknownPropertyError(Exception):
    """
    Raised, when there is an attempt to create instance with unknown property
    provided.
    """


class LMIMethodCallError(Exception):
    """
    Raised, when an error occurs within method call.
    """


class LMISynchroMethodCallError(Exception):
    """
    Raised, when an error occurs within synchronized method call.
    """


class LMISynchroMethodCallFilterError(Exception):
    """
    Raised, when the LMIShell can not find necessary static filter for
    synchronous method call.
    """


class LMINoPagerError(Exception):
    """
    Raised, when there is no default pager like less or more.
    """


class LMIIndicationError(Exception):
    """
    Raised, if an error occurs while subscribing to/removing an indication.
    """


class LMIIndicationListenerError(Exception):
    """
    Raised, if there is an error while starting/stopping indication listener.
    """


class LMIHandlerNamePatternError(Exception):
    """
    Raised when the pattern string does not contain minimum replaceable
    characters.
    """


class LMIFilterError(Exception):
    """
    Raised, when a filter error occurs, mostly when filter object is missing.
    """


class LMIClassNotFound(AttributeError):
    """
    Raised, when trying to access missing class in LMINamespace.

    :param string namespace: namespace name
    :param string classname: class name, which was not found in **namespace**
    """
    def __init__(self, namespace, class_name):
        AttributeError.__init__(
            self, 'no such class "%s" in "%s" namespace' %
            (class_name, namespace))


class LMINamespaceNotFound(AttributeError):
    """
    Raised, when trying to access not existing namespace from connection or
    namespace object.

    :param string namespace: namespace which was not found
    :param args: other positional arguments
    """
    def __init__(self, namespace, *args):
        AttributeError.__init__(self, 'no such namespace "%s"' %
                                "/".join([namespace] + list(args)))


class LMINotSupported(Exception):
    """
    Raised, when non-WSMAN method is about to be called.
    """


class CIMError(Exception):
    """
    LMIShell's exception for CIM errors.
    """
    def __init__(self, *args):
        super(CIMError, self).__init__()
        if len(args) == 1 and isinstance(args[0], Exception):
            self.args = args[0].args
        else:
            self.args = args

class ConnectionError(Exception):
    """
    LMIShell's exception for Connection errors.
    """
    def __init__(self, *args):
        super(ConnectionError, self).__init__()
        if len(args) == 1 and isinstance(args[0], Exception):
            self.args = args[0].args
        else:
            self.args = args



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Meta-command utility module.
"""

import logging
import logging.config
import os
import pkg_resources
import platform
import re
import socket
import sys
import urlparse

from lmi.scripts.common import Configuration
from lmi.scripts.common import get_logger
from lmi.scripts.common import lmi_logging

PYTHON_EGG_NAME = "openlmi-tools"
#: Service name identifying tcp port used to connect to CIMOM.
DEFAULT_BROKER_SERVICE_NAME = 'wbem-https'

RE_NETLOC = re.compile(r'^((?P<username>[^:@]+)(:(?P<password>[^@]+))?@)?'
        r'(?P<hostname>[^:]+)(:(?P<port>\d+))?$')

VERBOSITY_2_LOG_LEVEL = {
    Configuration.OUTPUT_SILENT  : logging.ERROR,
    Configuration.OUTPUT_WARNING : logging.WARNING,
    Configuration.OUTPUT_INFO    : logging.INFO,
    Configuration.OUTPUT_DEBUG   : logging.DEBUG,
}

DEFAULT_LOGGING_CONFIG = {
    'version' : 1,
    'disable_existing_loggers': True,
    'formatters' : {
        'console' : {
            "()" : "lmi.scripts.common.lmi_logging.LevelDispatchingFormatter",
            "formatters" : {
                logging.INFO :
                    Configuration.default_options()['ConsoleInfoFormat']
            },
            'default': Configuration.default_options()['ConsoleFormat'],
            'datefmt' : '%Y-%m-%d %H:%M:%S'
        },
        'file' : {
            'format' : Configuration.default_options()['FileFormat'],
            'datefmt' : '%Y-%m-%d %H:%M:%S'
        }
    },
    'handlers' : {
        'console': {
            'class' : "logging.StreamHandler",
            'level' : logging.ERROR,
            'formatter': 'console',
        },
        'console_shell': {
            'class' : "logging.StreamHandler",
            'level' : logging.CRITICAL,
            'formatter': 'console',
        },
        'file' : {
            'class' : "logging.FileHandler",
            'level' : Configuration.default_options()['Level'].upper(),
            'formatter': 'file',
        }
    },
    'loggers' : {
        'lmi.shell' : {
            'handlers' : ['console_shell'],
            'level' : logging.DEBUG,
            'propagate' : False
        }
    },
    'root' : {
        'level' : logging.DEBUG,
        'handlers' : ['console']
    }
}

LOG = get_logger(__name__)

def setup_logging(app_config, stderr=sys.stderr):
    """
    Setup logging to console and optionally to the file.

    :param app_config: (``lmi.scripts.common.Configuration``)
        Configuration object.
    :param stderr: (``file``) Output stream, where console handler should
        dispatch logging messages.
    """
    cfg = DEFAULT_LOGGING_CONFIG.copy()

    # Set up logging to a file
    if app_config.log_file:
        cfg['handlers']['file']['filename'] = app_config.log_file
        cfg['formatters']['file']['format'] = app_config.get_safe(
                'Log', 'FileFormat', raw=True)
        try:
            cfg['handlers']['file']['level'] = \
                    getattr(logging, app_config.logging_level.upper())
        except KeyError:
            LOG().error('Unsupported logging level: "%s".',
                    app_config.logging_level)
        cfg['root']['handlers'].append('file')
        cfg['loggers']['lmi.shell']['handlers'].append('file')
    else:
        del cfg['formatters']['file']
        del cfg['handlers']['file']

    if app_config.get_safe('Log', "LogToConsole", bool):
        # Set up logging to console
        if stderr is not sys.stderr:
            cfg['handlers']['console']['stream'] = stderr
        cfg['handlers']['console']['level'] = \
                VERBOSITY_2_LOG_LEVEL[app_config.verbosity]

        # make the verbosity of lmi shell one level less
        lmi_shell_verbosity = app_config.verbosity
        if (   app_config.verbosity < app_config.OUTPUT_DEBUG
           and app_config.verbosity > app_config.OUTPUT_SILENT):
            lmi_shell_verbosity -= 1
        cfg['handlers']['console_shell']['level'] = \
                VERBOSITY_2_LOG_LEVEL[lmi_shell_verbosity]

        # use ConsoleInfoFormat for INFO and less severe levels
        cfg['formatters']['console']['formatters'] = {
            logging.INFO :
                    app_config.get_safe('Log', 'ConsoleInfoFormat', raw=True)
        }
        # use ConsoleFormat for any other level
        cfg['formatters']['console']['default'] = app_config.get_safe(
                'Log', 'ConsoleFormat', raw=True)
    else:
        cfg['root']['handlers'].remove('console')
        cfg['loggers']['lmi.shell']['handlers'].remove('console_shell')
        del cfg['handlers']['console']
        del cfg['handlers']['console_shell']

    use_colors = platform.system() != 'Windows' and stderr.isatty()
    lmi_logging.setup_logger(use_colors = use_colors)
    logging.config.dictConfig(cfg)

def get_version(egg_name=PYTHON_EGG_NAME):
    """
    Gets version string of any python egg. Defaults to the egg of current
    application.
    """
    return pkg_resources.get_distribution(egg_name).version

def get_hosts_credentials(hostnames):
    """
    Parse list of hostnames, get credentials out of them and return
    ``(hostnames, creds)``, where ``hostnames`` is a list of ``hostnames``
    with credentials removed and ``creds`` is a dictionary with a pair
    ``(username, password)`` for every hostname, that supplied it.

    :param hostnames: (``list``) List of hostnames with optional credentials.
        For example: ``http://root:password@hostname:5988``.
    """
    if not hasattr(hostnames, '__iter__'):
        raise TypeError("hostnames must be a list of hosts")
    new_hostnames = []
    credentials = {}
    for hostname in hostnames:
        parsed = urlparse.urlparse(hostname)
        if not parsed.netloc and parsed.path:
            # got something like [user[:pass]@]hostname[:port] (no scheme)
            match = RE_NETLOC.match(hostname)
            if match:
                hostname = match.group('hostname')
                if match.group('port'):
                    hostname += ':' + match.group('port')
                if match.group('username') or match.group('password'):
                    credentials[hostname] = (
                        match.group('username'), match.group('password'))
        elif parsed.username or parsed.password:
            hostname = parsed.scheme
            if parsed.scheme:
                hostname += "://"
            hostname += parsed.hostname
            if parsed.port:
                hostname += ":" + str(parsed.port)
            hostname += parsed.path
            credentials[hostname] = (parsed.username, parsed.password)
        new_hostnames.append(hostname)
    return (new_hostnames, credentials)

def parse_hosts_file(hosts_file):
    """
    Parse file with hostnames to connect to. Return list of parsed hostnames.

    :param hosts_file: (``file``) File object openned for read.
        It containes hostnames. Each hostname occupies single line.
    :rtype: (``tuple``) A pair of ``(hosts, creds)``, where ``hosts`` is a list
        of string with hostnames and ``creds`` is a dictionary mapping
        ``(username, password)`` to each hostname if supplied.
    """
    hostnames = []
    for line in hosts_file.readlines():
        hostnames.append(line.strip())
    return get_hosts_credentials(hostnames)

def get_default_hostname(port=None):
    """
    Choose default hostname to connect to. If logged as root, this will default
    to localhost, which results in unix socket being used for connection.
    Otherwise use full qualified domain name and hostname will be tried in this
    order. If they are not address-resolvable, '127.0.0.1' is returned.

    This function shall be used only if no uri is specified on command line.

    :param port: Port of desired service running on host (CIMOM broker).
        This defaults to :py:attr:`DEFAULT_BROKER_SERVICE_NAME`
    :type port: string or int
    """
    if port is None:
        port = DEFAULT_BROKER_SERVICE_NAME
    # Functions used to get hostname. first resolvable result will be used.
    name_getters = []
    if os.getuid() != 0:
        # Use non-'localhost' name only if we'are not logged in as root
        # for it prevents the use of unix socket.
        name_getters.extend([socket.getfqdn, socket.gethostname])
    name_getters.append(lambda: 'localhost')
    for name_func in name_getters:
        try:
            hostname = name_func()
            socket.getaddrinfo(hostname, port)
            break
        except socket.gaierror:
            pass
    else:
        hostname = '127.0.0.1'
    return hostname



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Contains command classes producing key-value pairs to output.
"""
import abc

from lmi.scripts.common import formatter
from lmi.scripts.common import get_logger
from lmi.scripts.common.command import meta
from lmi.scripts.common.command.session import LmiSessionCommand

LOG = get_logger(__name__)

class LmiShowInstance(LmiSessionCommand):
    """
    End point command producing a list of properties of particular CIM
    instance. Either reduced list of properties to print can be specified, or
    the associated function alone can decide, which properties shall be
    printed. Associated function is expected to return CIM instance as a
    result.

    List of additional recognized properties:

        ``DYNAMIC_PROPERTIES`` : ``bool``
            A boolean saying, whether the associated function alone shall
            specify the list of properties of rendered instance. If ``True``,
            the result of function must be a pair: ::

                (props, inst)

            Where props is the same value as can be passed to ``PROPERTIES``
            property. Defaults to ``False``.
        ``PROPERTIES`` : ``tuple``
            May contain list of instance properties, that will be produced in
            the same order as output. Each item of list can be either:

                name : ``str``
                    Name of property to render.
                pair : ``tuple``
                    A tuple ``(Name, render_func)``, where former item an
                    arbitraty name for rendered value and the latter is a
                    function taking as the only argument particular instance
                    and returning value to render.

    ``DYNAMIC_PROPERTIES`` and ``PROPERTIES`` are mutually exclusive. If none
    is given, all instance properties will be printed.

    Using metaclass: :py:class:`~.meta.ShowInstanceMetaClass`.
    """
    __metaclass__ = meta.ShowInstanceMetaClass

    def formatter_factory(self):
        return formatter.SingleFormatter

    @abc.abstractmethod
    def render(self, result):
        """
        This method can either be overriden in a subclass or left alone. In the
        latter case it will be generated by
        :py:class:`~.meta.ShowInstanceMetaClass` metaclass with regard to
        ``PROPERTIES`` and ``DYNAMIC_PROPERTIES``.

        :param result: Either an instance to
            render or pair of properties and instance.
        :type: :py:class:`lmi.shell.LMIInstance` or ``tuple``
        :returns: List of pairs, where the first item is a label and second a
            value to render.
        :rtype: list
        """
        raise NotImplementedError(
                "render method must be overriden in subclass")

    def take_action(self, connection, args, kwargs):
        """
        Process single connection to a host, render result and return a value
        to render.

        :returns: List of pairs, where the first item is a label and
            second a value to render.
        :rtype: list
        """
        res = self.execute_on_connection(connection, *args, **kwargs)
        return self.render(res)




# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.


class LMISubscription(object):
    """
    Class holding information about a indication subscription.

    :param LMIShellClient client: client object used for CIMOM communication
    :param tuple cim_filter: contains filter object and bool indicator,
        if the filter object was created temporarily
    :param tuple cim_handler: contains handler object and bool indicator,
        if the handler object was created temporarily
    :param cim_subscription: subscription object
    :param bool permanent: indicates, if the indication should be deleted on
        the LMIShell's quit
    """
    def __init__(self, client, cim_filter, cim_handler, cim_subscription,
                 permanent):
        self._client = client
        self._cim_filter_tpl = cim_filter
        self._cim_handler_tpl = cim_handler
        self._cim_subscription = cim_subscription
        self._permanent = permanent

    def delete(self):
        """
        Cleans up the indication subscription.

        First it deletes subscription object. If
        :py:attr:`.LMISubscription._cim_filter_tpl` contains a flag, that the
        filter object was created temporarily, it will be deleted by this call.
        If :py:attr:`.LMISubscription._cim_handler_tlp` contains an flag, that
        the handler object was created temporarily, it will be deleted as well.

        This is called from :py:class:`.LMIConnection` object, which holds an
        internal list of all subscribed indications by the LMIShell (if not
        created by hand).
        """
        self._client.delete_instance(self._cim_subscription.path)
        if self._cim_filter_tpl[1]:
            self._client.delete_instance(self._cim_filter_tpl[0].path)
        if self._cim_handler_tpl[1]:
            self._client.delete_instance(self._cim_handler_tpl[0].path)

    @property
    def permanent(self):
        return self._permanent



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import sys


class LMIShellConfig(object):
    """
    Class representing the shell's configuration. The class is responsible for
    loading the configuration from the file and provides a unified API to
    access these settings.

    Constructs a :py:class:`.LMIShellConfig` object and loads the configuration
    from the file. If there is no such file, the shell's configuration
    properties are set to default values. Configuration file uses python
    syntax. If there is a syntax error in the configuration file, the
    properties are set to default values, as well.
    """
    DEFAULT_CONFIG_FILE = "~/.lmishellrc"
    DEFAULT_HISTORY_FILE = "~/.lmishell_history"
    DEFAULT_HISTORY_LENGTH = -1
    DEFAULT_USE_CACHE = True
    DEFAULT_USE_EXCEPTIONS = False
    DEFAULT_LISTENER_CERT_FILE = ""
    DEFAULT_LISTENER_KEY_FILE = ""

    def __init__(self):
        try:
            conf = {}
            execfile(os.path.expanduser(
                LMIShellConfig.DEFAULT_CONFIG_FILE), conf)
            self._history_file = os.path.expanduser(
                conf.get("history_file",
                         LMIShellConfig.DEFAULT_HISTORY_FILE))
            self._history_length = conf.get(
                "history_length", LMIShellConfig.DEFAULT_HISTORY_LENGTH)
            self._use_cache = conf.get(
                "use_cache", LMIShellConfig.DEFAULT_USE_CACHE)
            self._use_exceptions = conf.get(
                "use_exceptions", LMIShellConfig.DEFAULT_USE_EXCEPTIONS)
            self._indication_cert_file = conf.get(
                "indication_cert_file",
                LMIShellConfig.DEFAULT_LISTENER_CERT_FILE)
            self._indication_key_file = conf.get(
                "indication_key_file",
                LMIShellConfig.DEFAULT_LISTENER_KEY_FILE)
        except (SyntaxError, IOError), e:
            if isinstance(e, SyntaxError):
                sys.stderr.write("Error: %s\n" % e)
            self._history_file = os.path.expanduser(
                LMIShellConfig.DEFAULT_HISTORY_FILE)
            self._history_length = LMIShellConfig.DEFAULT_HISTORY_LENGTH
            self._use_cache = LMIShellConfig.DEFAULT_USE_CACHE
            self._use_exceptions = LMIShellConfig.DEFAULT_USE_EXCEPTIONS
            self._indication_cert_file = \
                LMIShellConfig.DEFAULT_LISTENER_CERT_FILE
            self._indication_key_file = \
                LMIShellConfig.DEFAULT_LISTENER_KEY_FILE

    @property
    def history_file(self):
        """
        Property returning a string containing the shell's history file.

        :returns: history file
        :rtype: string
        """
        return self._history_file

    @property
    def history_length(self):
        """
        Property returning a number with the shell's history file length.

        :returns: history file length
        :rtype: int
        """
        return self._history_length

    @property
    def use_cache(self):
        """
        Property returning a bool flag, if the shell should use cache for class
        retrieval.

        :returns: flag, if the shell should use a cache
        :rtype: bool
        """
        return self._use_cache

    @property
    def use_exceptions(self):
        """
        Property returning a bool flag, if the shell should throw the
        exceptions away, or if they should be propagated further.

        :returns: flag, if the shell should use exceptions, or throw them away
        :rtype: bool
        """
        return self._use_exceptions

    @property
    def cert_file(self):
        """
        Property returning a file name containing x509 certificate. This
        is used for :py:class:`.LMIIndicationListener`.

        :returns: x509 certificate file name
        :rtype: string
        """
        return self._indication_cert_file

    @property
    def key_file(self):
        """
        Property returning a file name containing x509 certificate private key.
        This is used for :py:class:`.LMIIndicationListener`.

        :returns: x509 certificate private key
        :rtype: string
        """
        return self._indication_key_file



# Copyright (C) 2012-2014 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import sys
import code
import readline
import __builtin__

from lmi.shell.LMIUtil import LMIPassByRef
from lmi.shell.LMIIndicationListener import LMIIndicationListener
from lmi.shell.LMICIMXMLClient import LMICIMXMLClient
from lmi.shell.LMIWSMANClient import LMIWSMANClient
from lmi.shell.LMIShellClient import LMIShellClient
from lmi.shell.LMIShellConfig import LMIShellConfig
from lmi.shell.LMIShellOptions import LMIShellOptions
from lmi.shell.LMISubscription import LMISubscription
from lmi.shell.LMICompleter import LMICompleter
from lmi.shell.LMIConnection import LMIConnection
from lmi.shell.LMIShellConfig import LMIShellConfig
from lmi.shell.LMIReturnValue import LMIReturnValue
from lmi.shell.LMIObjectFactory import LMIObjectFactory

from lmi.shell.LMIConnection import connect as connect_internal

from lmi.shell.LMIExceptions import CIMError
from lmi.shell.LMIExceptions import ConnectionError

from lmi.shell.LMIUtil import lmi_set_use_exceptions
from lmi.shell.LMIUtil import lmi_get_use_exceptions
from lmi.shell.LMIUtil import lmi_associators
from lmi.shell.LMIUtil import lmi_isinstance

from lmi.shell.LMIHelper import LMIHelper


class LMIConsole(code.InteractiveConsole):
    """
    Class representing an interactive console.
    """
    SHELL_PS1 = "> "
    SHELL_PS2 = "... "
    DEFAULT_LOCALS = {
        "__name__": "__main__",
        "LMIPassByRef": LMIPassByRef,
        "LMIIndicationListener": LMIIndicationListener,
        "LMIReturnValue": LMIReturnValue,
        "LMICIMXMLClient": LMICIMXMLClient,
        "LMIWSMANClient": LMIWSMANClient,
        "LMIShellClient": LMIShellClient,
        "LMIShellConfig": LMIShellConfig,
        "LMIShellOptions": LMIShellOptions,
        "LMISubscription": LMISubscription,
        "LMIConnection": LMIConnection,
        "LMINamespace": LMIObjectFactory().LMINamespace,
        "LMINamespaceRoot": LMIObjectFactory().LMINamespaceRoot,
        "LMIClass": LMIObjectFactory().LMIClass,
        "LMIInstance": LMIObjectFactory().LMIInstance,
        "LMIInstanceName": LMIObjectFactory().LMIInstanceName,
        "LMIMethod": LMIObjectFactory().LMIMethod,
        "CIMError": CIMError,
        "ConnectionError": ConnectionError,
        "use_exceptions": lmi_set_use_exceptions,
        "lmi_associators": lmi_associators,
        "lmi_isinstance": lmi_isinstance,
        "help": LMIHelper()
    }

    def __init__(self, cwd_first_in_path=False):
        # Setup default LMIConsole locals, that are needed
        # for proper object construction
        locals = dict(LMIConsole.DEFAULT_LOCALS.items())
        code.InteractiveConsole.__init__(self, locals)

        # Store configuration
        config = LMIShellConfig()
        self._history_file = config.history_file
        self._history_length = config.history_length
        self._use_cache = config.use_cache
        self._use_exceptions = config.use_exceptions
        self._verify_server_cert = True

        # Setup LMIShell-wide option, which defines, whether the LMIShell
        # should propagate caught exceptions, or dump them
        lmi_set_use_exceptions(self._use_exceptions)

        # Append/Prepend current working directory into Python's search list,
        # so we can properly import any module placed in CWD. By default, the
        # CWD is appended for security reasons.
        if cwd_first_in_path:
            sys.path.insert(0, os.getcwd())
        else:
            sys.path.append(os.getcwd())

    def setup_completer(self):
        """
        Initializes tab-completer.
        """
        self._completer = LMICompleter(self.locals)
        readline.set_completer(self._completer.complete)
        readline.parse_and_bind("tab: complete")

    def interact(self, locals=None):
        """
        Starts the interactive mode.

        :param dictionary locals: locals
        """
        def connect(uri, username="", password="", key_file=None,
                    cert_file=None,
                    verify_server_cert=self._verify_server_cert,
                    prompt_prefix=""):
            """
            Returns :py:class:`.LMIConnection` object with provided URI and
            credentials.

            :param string uri: URI of the CIMOM
            :param string username: account, under which, the CIM calls will be
                performed
            :param string password: user's password
            :param dictionary kwargs: supported keyword arguments
            :param string key_file: path to x509 key file; default value is
                None
            :param string cert_file: path to x509 cert file; default value is
                None
            :param bool verify_server_cert** (*bool*) -- tells, if the server
                side certificate needs to be verified, if SSL used; default
                value is True
            :param string prompt_prefix: used for username/password prompt;
                default value is empty string

            :returns: connection object
            :rtype: :py:class:`.LMIConnection`
            """
            return connect_internal(
                uri, username, password, interactive=True,
                use_cache=self._use_cache, key_file=key_file,
                cert_file=cert_file, verify_server_cert=verify_server_cert,
                prompt_prefix=prompt_prefix)

        # Initialize the interpreter
        if locals is None:
            locals = {}
        else:
            for k, v in locals.iteritems():
                if isinstance(v, LMIConnection):
                    locals[k].client.interactive = True
        locals["clear_history"] = self.clear_history
        locals["connect"] = connect
        self.locals = dict(self.locals.items() + locals.items())
        self.setup_completer()
        self.load_history()
        old_sys_ps1 = ""
        try:
            old_sys_ps1 = sys.ps1
        except AttributeError, e:
            old_sys_ps1 = ""
        try:
            old_sys_ps2 = sys.ps2
        except AttributeError, e:
            old_sys_ps2 = ""
        sys.ps1 = LMIConsole.SHELL_PS1
        sys.ps2 = LMIConsole.SHELL_PS2

        # Interact
        more = 0
        while 1:
            try:
                if more:
                    prompt = sys.ps2
                else:
                    prompt = sys.ps1
                try:
                    line = self.raw_input(prompt)
                    # Can be None if sys.stdin was redefined
                    encoding = getattr(sys.stdin, "encoding", None)
                    if encoding and not isinstance(line, unicode):
                        line = line.decode(encoding)
                except EOFError:
                    self.write("\n")
                    break
                else:
                    more = self.push(line)
            except KeyboardInterrupt:
                self.write("\n")
                self.resetbuffer()
                more = 0

        # Cleanup after the interpreter
        if old_sys_ps1:
            sys.ps1 = old_sys_ps1
        if old_sys_ps2:
            sys.ps2 = old_sys_ps2
        self.save_history()

    def interpret(self, script_name, script_argv, locals=None,
                  interactive=False):
        """
        Interprets a specified script within additional provided locals.
        There are :py:attr:`LMIConsole.DEFAULT_LOCALS` present.

        :param string script_name: script name
        :param list script_argv: script CLI arguments
        :param dictionary locals: dictionary with locals
        :param bool interactive: tells LMIShell, if the script should be
            treated as if it was run in interactive mode
        :returns: exit code of the script
        :rtype: int
        """
        # Helper function for LMIShell's scripts
        def connect(uri, username="", password="", key_file=None,
                    cert_file=None,
                    verify_server_cert=self._verify_server_cert,
                    prompt_prefix=""):
            """
            Returns :py:class:`.LMIConnection` object with provided URI and
            credentials.

            :param string uri: URI of the CIMOM
            :param string username: account, under which, the CIM calls will be
                performed
            :param string password: user's password
            :param dictionary kwargs: supported keyword arguments
            :param string key_file: path to x509 key file; default value is
                None
            :param string cert_file: path to x509 cert file; default value is
                None
            :param bool verify_server_cert: tells, if the server side
                certificate needs to be verified, if SSL used; default value is
                True
            :param string prompt_prefix: used for username/password prompt;
                default value is empty string

            :returns: connection object
            :rtype: :py:class:`.LMIConnection`
            """
            return connect_internal(
                uri, username, password, interactive=interactive,
                use_cache=self._use_cache, key_file=key_file,
                cert_file=cert_file, verify_server_cert=verify_server_cert,
                prompt_prefix=prompt_prefix)
        # Initialize locals
        if locals is None:
            locals = {}
        else:
            for k, v in locals.iteritems():
                if isinstance(v, LMIConnection):
                    locals[k].client.interactive = False
        locals["connect"] = connect
        self.locals = dict(self.locals.items() + locals.items())

        # Execute the script
        sys.argv = script_argv
        exit_code = 0
        try:
            execfile(script_name, self.locals)
        except SystemExit, e:
            exit_code = e.code

        # Return with exit_code
        return exit_code

    def load_history(self):
        """
        Loads the shell's history from the history file.
        """
        if self._history_length == 0 or not os.path.exists(self._history_file):
            return
        readline.read_history_file(self._history_file)
        if self._history_length > 0 and \
                readline.get_current_history_length() > self._history_length:
            readline.set_history_length(self._history_length)
            readline.write_history_file(self._history_file)
            readline.read_history_file(self._history_file)

    def save_history(self):
        """
        Saves current history of commands into the history file. If the length
        of history exceeds a maximum history file length, the history will be
        truncated.
        """
        if self._history_length == 0:
            return
        elif self._history_length > 0:
            readline.set_history_length(self._history_length)
        try:
            readline.write_history_file(self._history_file)
        except IOError, e:
            pass

    def clear_history(self):
        """
        Clears the current history.
        """
        readline.clear_history()

    def set_verify_server_certificate(self, verify_server_cert=True):
        """
        Turns on or off server side certificate verification, if SSL used.

        :param bool verify_server_cert: -- flag which tells, whether a server
            side certificate needs to be verified, if SSL used
        """
        self._verify_server_cert = verify_server_cert



# Copyright (C) 2013-2014 Michal Minar <miminar@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
"""
Contains command classes used to control formatters from inside of command
execution functions.
"""

class FormatterCommand(object):
    """
    Base class for formatter commands.
    """
    pass

class NewHostCommand(FormatterCommand):
    """
    Command for formatter to finish current table (if any), print header for
    new host and (if there are any data) print table header.

    :param string hostname: Name of host appearing at the front of new table.
    """
    def __init__(self, hostname):
        self.hostname = hostname

class NewTableCommand(FormatterCommand):
    """
    Command for formatter to finish current table (if any), print the **title**
    and (if there are any data) print table header.

    :param string title: Optional title for new table.
    """
    def __init__(self, title=None):
        self.title = title

class NewTableHeaderCommand(FormatterCommand):
    """
    Command for formatter to finish current table (if any), store new table
    header and (if there are any data) print the table header.
    The table header will be printed in all subsequent tables, until
    new instance of this class arrives.

    :param tuple columns: Array of column names.
    """
    def __init__(self, columns=None):
        self.columns = columns
